# Copyright (c) 2023-2024 Geosiris.
# SPDX-License-Identifier: Apache-2.0
import inspect
import logging
import sys
from typing import Any, Optional, Callable, List, Union

from .datasets_io import read_external_dataset_array
from ..constants import flatten_concatenation
from ..epc import get_obj_identifier
from ..exception import ObjectNotFoundNotError
from ..introspection import (
    snake_case,
    get_object_attribute_no_verif,
    search_attribute_matching_name_with_path,
    search_attribute_matching_name,
    search_attribute_in_upper_matching_name,
    get_obj_uuid,
    get_object_attribute,
    get_object_attribute_rgx,
)
from ..workspace import EnergymlWorkspace

_ARRAY_NAMES_ = [
    "BooleanArrayFromDiscretePropertyArray",
    "BooleanArrayFromIndexArray",
    "BooleanConstantArray",
    "BooleanExternalArray",
    "BooleanHdf5Array",
    "BooleanXmlArray",
    "CompoundExternalArray",
    "DasTimeArray",
    "DoubleConstantArray",
    "DoubleHdf5Array",
    "DoubleLatticeArray",
    "ExternalDataArray",
    "FloatingPointConstantArray",
    "FloatingPointExternalArray",
    "FloatingPointLatticeArray",
    "FloatingPointXmlArray",
    "IntegerArrayFromBooleanMaskArray",
    "IntegerConstantArray",
    "IntegerExternalArray",
    "IntegerHdf5Array",
    "IntegerLatticeArray",
    "IntegerRangeArray",
    "IntegerXmlArray",
    "JaggedArray",
    "ParametricLineArray",
    "ParametricLineFromRepresentationLatticeArray",
    "Point2DHdf5Array",
    "Point3DFromRepresentationLatticeArray",
    "Point3DHdf5Array",
    "Point3DLatticeArray",
    "Point3DParametricArray",
    "Point3DZvalueArray",
    "ResqmlJaggedArray",
    "StringConstantArray",
    "StringExternalArray",
    "StringHdf5Array",
    "StringXmlArray",
]


def _point_as_array(point: Any) -> List:
    """
    Transform a point that has "coordinate1", "coordinate2", "coordinate3" as attributes into a list.
    :param point:
    :return:
    """
    return [
        get_object_attribute_no_verif(point, "coordinate1"),
        get_object_attribute_no_verif(point, "coordinate2"),
        get_object_attribute_no_verif(point, "coordinate3"),
    ]


def is_z_reversed(crs: Optional[Any]) -> bool:
    """
    Returns True if the Z axe is reverse (ZIncreasingDownward=='True' or VerticalAxis.Direction=='down')
    :param crs: a CRS object
    :return: By default, False is returned (if 'crs' is None)
    """
    reverse_z_values = False
    if crs is not None:
        # resqml 201
        zincreasing_downward = search_attribute_matching_name(crs, "ZIncreasingDownward")
        if len(zincreasing_downward) > 0:
            reverse_z_values = zincreasing_downward[0]

        # resqml >= 22
        vert_axis = search_attribute_matching_name(crs, "VerticalAxis.Direction")
        if len(vert_axis) > 0:
            vert_axis_str = str(vert_axis[0])
            if "." in vert_axis_str:
                vert_axis_str = vert_axis_str.split(".")[-1]

            reverse_z_values = vert_axis_str.lower() == "down"

    return reverse_z_values


def get_vertical_epsg_code(crs_object: Any):
    vertical_epsg_code = None
    if crs_object is not None:  # LocalDepth3dCRS
        vertical_epsg_code = get_object_attribute_rgx(crs_object, "VerticalCrs.EpsgCode")
        if vertical_epsg_code is None:  # LocalEngineering2DCrs
            vertical_epsg_code = get_object_attribute_rgx(
                crs_object, "OriginProjectedCrs.AbstractProjectedCrs.EpsgCode"
            )
    return vertical_epsg_code


def get_projected_epsg_code(crs_object: Any, workspace: Optional[EnergymlWorkspace] = None):
    if crs_object is not None:  # LocalDepth3dCRS
        projected_epsg_code = get_object_attribute_rgx(crs_object, "ProjectedCrs.EpsgCode")
        if projected_epsg_code is None:  # LocalEngineering2DCrs
            projected_epsg_code = get_object_attribute_rgx(
                crs_object, "OriginProjectedCrs.AbstractProjectedCrs.EpsgCode"
            )

        if projected_epsg_code is None and workspace is not None:
            return get_projected_epsg_code(
                workspace.get_object_by_uuid(get_object_attribute_rgx(crs_object, "LocalEngineering2[dD]Crs.Uuid"))
            )
        return projected_epsg_code
    return None


def get_projected_uom(crs_object: Any, workspace: Optional[EnergymlWorkspace] = None):
    if crs_object is not None:
        projected_epsg_uom = get_object_attribute_rgx(crs_object, "ProjectedUom")
        if projected_epsg_uom is None:
            projected_epsg_uom = get_object_attribute_rgx(crs_object, "HorizontalAxes.ProjectedUom")

        if projected_epsg_uom is None and workspace is not None:
            return get_projected_uom(
                workspace.get_object_by_uuid(get_object_attribute_rgx(crs_object, "LocalEngineering2[dD]Crs.Uuid"))
            )
        return projected_epsg_uom
    return None


def get_crs_origin_offset(crs_obj: Any) -> List[float]:
    """
    Return a list [X,Y,Z] corresponding to the crs Offset [XOffset/OriginProjectedCoordinate1, ... ] depending on the
    crs energyml version.
    :param crs_obj:
    :return:
    """
    tmp_offset_x = get_object_attribute_rgx(crs_obj, "XOffset")
    if tmp_offset_x is None:
        tmp_offset_x = get_object_attribute_rgx(crs_obj, "OriginProjectedCoordinate1")

    tmp_offset_y = get_object_attribute_rgx(crs_obj, "YOffset")
    if tmp_offset_y is None:
        tmp_offset_y = get_object_attribute_rgx(crs_obj, "OriginProjectedCoordinate2")

    tmp_offset_z = get_object_attribute_rgx(crs_obj, "ZOffset")
    if tmp_offset_z is None:
        tmp_offset_z = get_object_attribute_rgx(crs_obj, "OriginProjectedCoordinate3")

    crs_point_offset = [0, 0, 0]
    try:
        crs_point_offset = [
            float(tmp_offset_x) if tmp_offset_x is not None else 0,
            float(tmp_offset_y) if tmp_offset_y is not None else 0,
            float(tmp_offset_z) if tmp_offset_z is not None else 0,
        ]
    except Exception as e:
        logging.info(f"ERR reading crs offset {e}")

    return crs_point_offset


def prod_n_tab(val: Union[float, int, str], tab: List[Union[float, int, str]]):
    """
    Multiply every value of the list 'tab' by the constant 'val'
    :param val:
    :param tab:
    :return:
    """
    return list(map(lambda x: x * val, tab))


def sum_lists(l1: List, l2: List):
    """
    Sums 2 lists values.

    Example:
        [1,1,1] and [2,2,3,6] gives : [3,3,4,6]

    :param l1:
    :param l2:
    :return:
    """
    return [l1[i] + l2[i] for i in range(min(len(l1), len(l2)))] + max(l1, l2, key=len)[min(len(l1), len(l2)) :]


def get_crs_obj(
    context_obj: Any,
    path_in_root: Optional[str] = None,
    root_obj: Optional[Any] = None,
    workspace: Optional[EnergymlWorkspace] = None,
) -> Optional[Any]:
    """
    Search for the CRS object related to :param:`context_obj` into the :param:`workspace`
    :param context_obj:
    :param path_in_root:
    :param root_obj:
    :param workspace:
    :return:
    """
    if workspace is None:
        logging.error("@get_crs_obj no Epc file given")
    else:
        crs_list = search_attribute_matching_name(context_obj, r"\.*Crs", search_in_sub_obj=True, deep_search=False)
        if crs_list is not None and len(crs_list) > 0:
            # logging.debug(crs_list[0])
            crs = workspace.get_object_by_identifier(get_obj_identifier(crs_list[0]))
            if crs is None:
                crs = workspace.get_object_by_uuid(get_obj_uuid(crs_list[0]))
            if crs is None:
                logging.error(f"CRS {crs_list[0]} not found (or not read correctly)")
                raise ObjectNotFoundNotError(get_obj_identifier(crs_list[0]))
            if crs is not None:
                return crs

        if context_obj != root_obj:
            upper_path = path_in_root[: path_in_root.rindex(".")]
            if len(upper_path) > 0:
                return get_crs_obj(
                    context_obj=get_object_attribute(root_obj, upper_path),
                    path_in_root=upper_path,
                    root_obj=root_obj,
                    workspace=workspace,
                )

    return None


#     ___
#    /   |  ______________ ___  _______
#   / /| | / ___/ ___/ __ `/ / / / ___/
#  / ___ |/ /  / /  / /_/ / /_/ (__  )
# /_/  |_/_/  /_/   \__,_/\__, /____/
#                        /____/


def _array_name_mapping(array_type_name: str) -> str:
    """
    Transform the type name to match existing reader function
    :param array_type_name:
    :return:
    """
    array_type_name = array_type_name.replace("3D", "3d").replace("2D", "2d")
    if array_type_name.endswith("ConstantArray"):
        return "ConstantArray"
    elif "External" in array_type_name or "Hdf5" in array_type_name:
        return "ExternalArray"
    elif array_type_name.endswith("XmlArray"):
        return "XmlArray"
    elif "Jagged" in array_type_name:
        return "JaggedArray"
    elif "Lattice" in array_type_name:
        if "Integer" in array_type_name or "Double" in array_type_name:
            return "int_double_lattice_array"
    return array_type_name


def get_supported_array() -> List[str]:
    """
    Return a list of the supported arrays for the use of :py:func:`energyml.utils.data.helper.read_array` function.
    :return:
    """
    return [x for x in _ARRAY_NAMES_ if get_array_reader_function(_array_name_mapping(x)) is not None]


def get_not_supported_array():
    """
    Return a list of the NOT supported arrays for the use of :py:func:`energyml.utils.data.helper.read_array` function.
    :return:
    """
    return [x for x in _ARRAY_NAMES_ if get_array_reader_function(_array_name_mapping(x)) is None]


def read_external_array(
    energyml_array: Any,
    root_obj: Optional[Any] = None,
    path_in_root: Optional[str] = None,
    workspace: Optional[EnergymlWorkspace] = None,
    sub_indices: List[int] = None,
) -> List[Any]:
    """
    Read an external array (BooleanExternalArray, BooleanHdf5Array, DoubleHdf5Array, IntegerHdf5Array, StringExternalArray ...)
    :param energyml_array:
    :param root_obj:
    :param path_in_root:
    :param workspace:
    :return:
    """
    array = None
    if workspace is not None:
        array = workspace.read_external_array(
            energyml_array=energyml_array,
            root_obj=root_obj,
            path_in_root=path_in_root,
        )
    else:
        array = read_external_dataset_array(
            energyml_array=energyml_array,
            root_obj=root_obj,
            path_in_root=path_in_root,
        )

    if sub_indices is not None and len(sub_indices) > 0:
        res = []
        for idx in sub_indices:
            res.append(array[idx])
        array = res

    return array


def get_array_reader_function(array_type_name: str) -> Optional[Callable]:
    """
    Returns the name of the potential appropriate function to read an object with type is named :param array_type_name
    :param array_type_name: the initial type name
    :return:
    """
    for name, obj in inspect.getmembers(sys.modules[__name__]):
        if name == f"read_{snake_case(array_type_name)}":
            return obj
    return None


def read_array(
    energyml_array: Any,
    root_obj: Optional[Any] = None,
    path_in_root: Optional[str] = None,
    workspace: Optional[EnergymlWorkspace] = None,
    sub_indices: List[int] = None,
) -> List[Any]:
    """
    Read an array and return a list. The array is read depending on its type. see. :py:func:`energyml.utils.data.helper.get_supported_array`
    :param energyml_array:
    :param root_obj:
    :param path_in_root:
    :param workspace:
    :param sub_indices: for SubRepresentation
    :return:
    """
    if isinstance(energyml_array, list):
        return energyml_array
    array_type_name = _array_name_mapping(type(energyml_array).__name__)

    reader_func = get_array_reader_function(array_type_name)
    if reader_func is not None:
        return reader_func(
            energyml_array=energyml_array,
            root_obj=root_obj,
            path_in_root=path_in_root,
            workspace=workspace,
            sub_indices=sub_indices,
        )
    else:
        logging.error(f"Type {array_type_name} is not supported: function read_{snake_case(array_type_name)} not found")
        raise Exception(
            f"Type {array_type_name} is not supported\n\t{energyml_array}: \n\tfunction read_{snake_case(array_type_name)} not found"
        )


def read_constant_array(
    energyml_array: Any,
    root_obj: Optional[Any] = None,
    path_in_root: Optional[str] = None,
    workspace: Optional[EnergymlWorkspace] = None,
    sub_indices: List[int] = None,
) -> List[Any]:
    """
    Read a constant array ( BooleanConstantArray, DoubleConstantArray, FloatingPointConstantArray, IntegerConstantArray ...)
    :param energyml_array:
    :param root_obj:
    :param path_in_root:
    :param workspace:
    :param sub_indices:
    :return:
    """
    # logging.debug(f"Reading constant array\n\t{energyml_array}")

    value = get_object_attribute_no_verif(energyml_array, "value")
    count = (
        len(sub_indices)
        if sub_indices is not None and len(sub_indices) > 0
        else get_object_attribute_no_verif(energyml_array, "count")
    )

    # logging.debug(f"\tValue : {[value for i in range(0, count)]}")

    return [value] * count


def read_xml_array(
    energyml_array: Any,
    root_obj: Optional[Any] = None,
    path_in_root: Optional[str] = None,
    workspace: Optional[EnergymlWorkspace] = None,
    sub_indices: List[int] = None,
) -> List[Any]:
    """
    Read a xml array ( BooleanXmlArray, FloatingPointXmlArray, IntegerXmlArray, StringXmlArray ...)
    :param energyml_array:
    :param root_obj:
    :param path_in_root:
    :param workspace:
    :param sub_indices:
    :return:
    """
    values = get_object_attribute_no_verif(energyml_array, "values")
    # count = get_object_attribute_no_verif(energyml_array, "count_per_value")

    if sub_indices is not None and len(sub_indices) > 0:
        res = []
        for idx in sub_indices:
            res.append(values[idx])
        values = res
    return values


def read_jagged_array(
    energyml_array: Any,
    root_obj: Optional[Any] = None,
    path_in_root: Optional[str] = None,
    workspace: Optional[EnergymlWorkspace] = None,
    sub_indices: List[int] = None,
) -> List[Any]:
    """
    Read a jagged array
    :param energyml_array:
    :param root_obj:
    :param path_in_root:
    :param workspace:
    :param sub_indices:
    :return:
    """
    elements = read_array(
        energyml_array=get_object_attribute_no_verif(energyml_array, "elements"),
        root_obj=root_obj,
        path_in_root=path_in_root + ".elements",
        workspace=workspace,
    )
    cumulative_length = read_array(
        energyml_array=read_array(get_object_attribute_no_verif(energyml_array, "cumulative_length")),
        root_obj=root_obj,
        path_in_root=path_in_root + ".cumulative_length",
        workspace=workspace,
    )

    array = []
    previous = 0
    for cl in cumulative_length:
        array.append(elements[previous:cl])
        previous = cl

    if sub_indices is not None and len(sub_indices) > 0:
        res = []
        for idx in sub_indices:
            res.append(array[idx])
        array = res
    return array


def read_int_double_lattice_array(
    energyml_array: Any,
    root_obj: Optional[Any] = None,
    path_in_root: Optional[str] = None,
    workspace: Optional[EnergymlWorkspace] = None,
    sub_indices: List[int] = None,
):
    """
    Read DoubleLatticeArray or IntegerLatticeArray.
    :param energyml_array:
    :param root_obj:
    :param path_in_root:
    :param workspace:
    :param sub_indices:
    :return:
    """
    start_value = get_object_attribute_no_verif(energyml_array, "start_value")
    offset = get_object_attribute_no_verif(energyml_array, "offset")

    result = []

    # if len(offset) == 1:
    #     pass
    # elif len(offset) == 2:
    #     pass
    # else:
    raise Exception(f"{type(energyml_array)} read with an offset of length {len(offset)} is not supported")

    # return result


def read_point3d_zvalue_array(
    energyml_array: Any,
    root_obj: Optional[Any] = None,
    path_in_root: Optional[str] = None,
    workspace: Optional[EnergymlWorkspace] = None,
    sub_indices: List[int] = None,
):
    """
    Read a Point3D2ValueArray
    :param energyml_array:
    :param root_obj:
    :param path_in_root:
    :param workspace:
    :param sub_indices:
    :return:
    """
    supporting_geometry = get_object_attribute_no_verif(energyml_array, "supporting_geometry")
    sup_geom_array = read_array(
        energyml_array=supporting_geometry,
        root_obj=root_obj,
        path_in_root=path_in_root + ".SupportingGeometry",
        workspace=workspace,
        sub_indices=sub_indices,
    )

    zvalues = get_object_attribute_no_verif(energyml_array, "zvalues")
    zvalues_array = flatten_concatenation(
        read_array(
            energyml_array=zvalues,
            root_obj=root_obj,
            path_in_root=path_in_root + ".ZValues",
            workspace=workspace,
            sub_indices=sub_indices,
        )
    )

    count = 0

    for i in range(len(sup_geom_array)):
        try:
            sup_geom_array[i][2] = zvalues_array[i]
        except Exception as e:
            if count == 0:
                logging.error(e, f": {i} is out of bound of {len(zvalues_array)}")
                count = count + 1

    return sup_geom_array


def read_point3d_from_representation_lattice_array(
    energyml_array: Any,
    root_obj: Optional[Any] = None,
    path_in_root: Optional[str] = None,
    workspace: Optional[EnergymlWorkspace] = None,
    sub_indices: List[int] = None,
):
    """
    Read a Point3DFromRepresentationLatticeArray.

    Note: Only works for Grid2DRepresentation.

    :param energyml_array:
    :param root_obj:
    :param path_in_root:
    :param workspace:
    :param sub_indices:
    :return:
    """
    supporting_rep_identifier = get_obj_identifier(
        get_object_attribute_no_verif(energyml_array, "supporting_representation")
    )
    # logging.debug(f"energyml_array : {energyml_array}\n\t{supporting_rep_identifier}")
    supporting_rep = workspace.get_object_by_identifier(supporting_rep_identifier)

    # TODO chercher un pattern \.*patch\.*.[d]+ pour trouver le numero du patch dans le path_in_root puis lire le patch
    # logging.debug(f"path_in_root {path_in_root}")

    result = []
    if "grid2d" in str(type(supporting_rep)).lower():
        patch_path, patch = search_attribute_matching_name_with_path(supporting_rep, "Grid2dPatch")[0]
        points = read_grid2d_patch(
            patch=patch, grid2d=supporting_rep, path_in_root=patch_path, workspace=workspace, sub_indices=sub_indices
        )
        # TODO: take the points by there indices from the NodeIndicesOnSupportingRepresentation
        result = points

    else:
        raise Exception(f"Not supported type {type(energyml_array)} for object {type(root_obj)}")
    # pour trouver les infos qu'il faut
    return result


def read_grid2d_patch(
    patch: Any,
    grid2d: Optional[Any] = None,
    path_in_root: Optional[str] = None,
    workspace: Optional[EnergymlWorkspace] = None,
    sub_indices: List[int] = None,
) -> List:
    points_path, points_obj = search_attribute_matching_name_with_path(patch, "Geometry.Points")[0]

    return read_array(
        energyml_array=points_obj,
        root_obj=grid2d,
        path_in_root=path_in_root + "." + points_path,
        workspace=workspace,
        sub_indices=sub_indices,
    )


def read_point3d_lattice_array(
    energyml_array: Any,
    root_obj: Optional[Any] = None,
    path_in_root: Optional[str] = None,
    workspace: Optional[EnergymlWorkspace] = None,
    sub_indices: List[int] = None,
) -> List:
    """
    Read a Point3DLatticeArray.

    Note: If a CRS is found and its 'ZIncreasingDownward' is set to true or its

    :param energyml_array:
    :param root_obj:
    :param path_in_root:
    :param workspace:
    :param sub_indices:
    :return:
    """
    result = []
    origin = _point_as_array(get_object_attribute_no_verif(energyml_array, "origin"))
    offset = get_object_attribute_rgx(energyml_array, "offset|dimension")

    if len(offset) == 2:
        slowest = offset[0]
        fastest = offset[1]

        crs_sa_count = search_attribute_in_upper_matching_name(
            obj=energyml_array,
            name_rgx="slowestAxisCount",
            root_obj=root_obj,
            current_path=path_in_root,
        )

        crs_fa_count = search_attribute_in_upper_matching_name(
            obj=energyml_array,
            name_rgx="fastestAxisCount",
            root_obj=root_obj,
            current_path=path_in_root,
        )

        crs = None
        try:
            crs = get_crs_obj(
                context_obj=energyml_array,
                path_in_root=path_in_root,
                root_obj=root_obj,
                workspace=workspace,
            )
        except ObjectNotFoundNotError as e:
            logging.error("No CRS found, not able to check zIncreasingDownward")

        zincreasing_downward = is_z_reversed(crs)

        slowest_vec = _point_as_array(get_object_attribute_rgx(slowest, "offset|direction"))
        slowest_spacing = read_array(get_object_attribute_no_verif(slowest, "spacing"))
        slowest_table = list(map(lambda x: prod_n_tab(x, slowest_vec), slowest_spacing))

        fastest_vec = _point_as_array(get_object_attribute_rgx(fastest, "offset|direction"))
        fastest_spacing = read_array(get_object_attribute_no_verif(fastest, "spacing"))
        fastest_table = list(map(lambda x: prod_n_tab(x, fastest_vec), fastest_spacing))

        slowest_size = len(slowest_table)
        fastest_size = len(fastest_table)

        if len(crs_sa_count) > 0 and len(crs_fa_count) > 0:
            if (crs_sa_count[0] == fastest_size and crs_fa_count[0] == slowest_size) or (
                crs_sa_count[0] == fastest_size - 1 and crs_fa_count[0] == slowest_size - 1
            ):
                logging.debug("reversing order")
                # if offset were given in the wrong order
                tmp_table = slowest_table
                slowest_table = fastest_table
                fastest_table = tmp_table

                tmp_size = slowest_size
                slowest_size = fastest_size
                fastest_size = tmp_size
            else:
                slowest_size = crs_sa_count[0]
                fastest_size = crs_fa_count[0]

        for i in range(slowest_size):
            for j in range(fastest_size):
                previous_value = origin
                # to avoid a sum of the parts of the array at each iteration, I take the previous value in the same line
                # number i and add the fastest_table[j] value

                if j > 0:
                    if i > 0:
                        line_idx = i * fastest_size  # numero de ligne
                        previous_value = result[line_idx + j - 1]
                    else:
                        previous_value = result[j - 1]
                    if zincreasing_downward:
                        result.append(sum_lists(previous_value, slowest_table[i - 1]))
                    else:
                        result.append(sum_lists(previous_value, fastest_table[j - 1]))
                else:
                    if i > 0:
                        prev_line_idx = (i - 1) * fastest_size  # numero de ligne precedent
                        previous_value = result[prev_line_idx]
                        if zincreasing_downward:
                            result.append(sum_lists(previous_value, fastest_table[j - 1]))
                        else:
                            result.append(sum_lists(previous_value, slowest_table[i - 1]))
                    else:
                        result.append(previous_value)
    else:
        raise Exception(f"{type(energyml_array)} read with an offset of length {len(offset)} is not supported")

    if sub_indices is not None and len(sub_indices) > 0:
        res = []
        for idx in sub_indices:
            res.append(result[idx])
        result = res

    return result


# def read_boolean_constant_array(
#         energyml_array: Any,
#         root_obj: Optional[Any] = None,
#         path_in_root: Optional[str] = None,
#         workspace: Optional[EnergymlWorkspace] = None
# ):
#     logging.debug(energyml_array)



# Copyright (c) 2023-2024 Geosiris.
# SPDX-License-Identifier: Apache-2.0
from .constants import *


@dataclass(init=True, eq=True,)
class Uri:
    """
    A class to represent an ETP URI
    """

    dataspace: Optional[str] = field(default=None)
    domain: Optional[str] = field(default=None)
    domain_version: Optional[str] = field(default=None)
    object_type: Optional[str] = field(default=None)
    uuid: Optional[str] = field(default=None)
    version: Optional[str] = field(default=None)
    collection_domain: Optional[str] = field(default=None)
    collection_domain_version: Optional[str] = field(default=None)
    collection_domain_type: Optional[str] = field(default=None)
    query: Optional[str] = field(default=None)

    @classmethod
    def parse(cls, uri: str):
        m = re.match(URI_RGX, uri, re.IGNORECASE)
        if m is not None:
            res = Uri()
            res.dataspace = m.group(URI_RGX_GRP_DATASPACE)
            res.domain = m.group(URI_RGX_GRP_DOMAIN)
            if res.domain is not None and len(res.domain) <= 0:
                res.domain = None
            res.domain_version = m.group(URI_RGX_GRP_DOMAIN_VERSION)
            res.object_type = m.group(URI_RGX_GRP_OBJECT_TYPE)
            res.uuid = m.group(URI_RGX_GRP_UUID) or m.group(URI_RGX_GRP_UUID2)
            res.version = m.group(URI_RGX_GRP_VERSION)
            res.collection_domain = m.group(URI_RGX_GRP_COLLECTION_DOMAIN)
            res.collection_domain_version = m.group(URI_RGX_GRP_COLLECTION_DOMAIN_VERSION)
            res.collection_domain_type = m.group(URI_RGX_GRP_COLLECTION_TYPE)
            res.query = m.group(URI_RGX_GRP_QUERY)
            return res
        else:
            return None

    def is_dataspace_uri(self):
        return (
            self.domain is None
            and self.object_type is None
            and self.query is None
            and self.collection_domain_type is None
        )

    def is_object_uri(self):
        return (
            self.domain is not None
            and self.domain_version is not None
            and self.object_type is not None
            and self.uuid is not None
        )

    def get_qualified_type(self):
        return f"{self.domain}{self.domain_version}.{self.object_type}"

    def __str__(self):
        res = "eml:///"
        if self.dataspace is not None and len(self.dataspace) > 0:
            res += f"dataspace('{self.dataspace}')"
            if self.domain is not None:
                res += "/"
        if self.domain is not None and self.domain_version is not None:
            res += f"{self.domain}{self.domain_version}.{self.object_type}"
            if self.uuid is not None:
                res += "("
                if self.version is not None:
                    res += f"uuid={self.uuid},version='{self.version}'"
                else:
                    res += self.uuid
                res += ")"
        if self.collection_domain is not None and self.collection_domain_version:
            res += f"/{self.collection_domain}{self.collection_domain_version}"
            if self.collection_domain_type is not None:
                res += f".{self.collection_domain_type}"

        if self.query is not None:
            res += f"?{self.query}"

        return res


def parse_uri(uri: str) -> Uri:
    return Uri.parse(uri)



# Copyright (c) 2023-2024 Geosiris.
# SPDX-License-Identifier: Apache-2.0
from typing import Optional


class DetailedNotImplementedError(Exception):
    """Exception for not implemented functions"""

    def __init__(self, msg):
        super().__init__(msg)


class MissingExtraInstallation(DetailedNotImplementedError):
    """Exception for missing extra installation"""

    def __init__(self, extra_name):
        super().__init__(msg=f"Missing energml-utils extras installation '{extra_name}'")


class NoCrsError(Exception):
    pass


class ObjectNotFoundNotError(Exception):
    def __init__(self, obj_id):
        super().__init__(f"Object id: {obj_id}")


class UnknownTypeFromQualifiedType(Exception):
    def __init__(self, qt: Optional[str] = None):
        super().__init__(f"not matchable qualified type: {qt}")


class NotParsableType(Exception):
    def __init__(self, t: Optional[str] = None):
        super().__init__(f"type: {t}")


class UnparsableFile(Exception):
    def __init__(self, t: Optional[str] = None):
        super().__init__(f"File is not parsable for an EPC file. Please use RawFile class for non energyml files.")



# Copyright (c) 2023-2024 Geosiris.
# SPDX-License-Identifier: Apache-2.0

"""
The energyml.utils module.
It contains tools for energyml management.

Please check the following module (depending on your need):
    - energyml-opc
    - energyml-common2-0
    - energyml-common2-1
    - energyml-common2-2
    - energyml-common2-3
    - energyml-resqml2-0-1
    - energyml-resqml2-2-dev3
    - energyml-resqml2-2
    - energyml-witsml2-0
    - energyml-witsml2-1
    - energyml-prodml2-0
    - energyml-prodml2-2
"""



# Copyright (c) 2023-2024 Geosiris.
# SPDX-License-Identifier: Apache-2.0
import json
import logging
import traceback
from enum import Enum
from io import BytesIO
from typing import Optional, Any, Union, List, Dict, Callable, Type

import xsdata
from lxml import etree
from xsdata.exceptions import ParserError
from xsdata.formats.dataclass.context import XmlContext
from xsdata.formats.dataclass.models.generics import DerivedElement
from xsdata.formats.dataclass.parsers import XmlParser, JsonParser
from xsdata.formats.dataclass.parsers.config import ParserConfig
from xsdata.formats.dataclass.parsers.handlers import (
    LxmlEventHandler,
    XmlEventHandler,
)
from xsdata.formats.dataclass.serializers import JsonSerializer
from xsdata.formats.dataclass.serializers import XmlSerializer
from xsdata.formats.dataclass.serializers.config import SerializerConfig

from .exception import UnknownTypeFromQualifiedType, NotParsableType
from .introspection import (
    get_class_from_name,
    get_energyml_class_in_related_dev_pkg,
    get_class_from_content_type,
    get_qualified_type_from_class,
    get_class_fields,
    get_obj_identifier,
    is_primitive,
    search_attribute_matching_name,
    get_class_from_qualified_type,
    get_matching_class_attribute_name,
    is_enum,
)
from .xml import (
    get_class_name_from_xml,
    get_tree,
    get_xml_encoding,
    ENERGYML_NAMESPACES,
)


class JSON_VERSION(Enum):
    XSDATA = "XSDATA"
    OSDU_OFFICIAL = "OSDU_OFFICIAL"


def _read_energyml_xml_bytes_as_class(
    file: bytes,
    obj_class: Type,
    fail_on_unknown_properties=True,
    fail_on_unknown_attributes=True,
) -> Any:
    """
    Read a xml file into the instance of type :param:`obj_class`.
    :param file:
    :param obj_class:
    :return:
    """
    config = ParserConfig(
        fail_on_unknown_properties=fail_on_unknown_properties,
        fail_on_unknown_attributes=fail_on_unknown_attributes,
        # process_xinclude=True,
    )
    parser = XmlParser(config=config)
    try:
        return parser.from_bytes(file, obj_class)
    except ParserError as e:
        logging.error(f"Failed to parse file {file} as class {obj_class}")
        if len(e.args) > 0:
            if "unknown property" in e.args[0].lower():
                logging.error(e)
                logging.error(
                    "A property has not been found, please check if your 'xsi::type' values contains "
                    "the xml namespace (e.g. 'xsi:type=\"eml:VerticalCrsEpsgCode\"')."
                )
        raise e


def read_energyml_xml_tree(file: etree, obj_type: Optional[type] = None) -> Any:
    # if obj_type is None:
    #     obj_type = get_class_from_name(get_class_name_from_xml(file))
    # parser = XmlParser(handler=XmlEventHandler)
    # # parser = XmlParser(handler=LxmlEventHandler)
    # return parser.parse(file, obj_type)
    return read_energyml_xml_bytes(etree.tostring(file, encoding="utf8"))


def read_energyml_xml_bytes(file: bytes, obj_type: Optional[type] = None) -> Any:
    """
    Read a xml file. The type of object is searched from the xml root name if not given.
    :param obj_type:
    :param file:
    :return:
    """
    if obj_type is None:
        obj_type = get_class_from_name(get_class_name_from_xml(get_tree(file)))
    try:
        return _read_energyml_xml_bytes_as_class(file, obj_type)
    except xsdata.exceptions.ParserError as e:
        if len(e.args) > 0:
            if "unknown property" in e.args[0].lower():
                logging.error(f"Trying reading without fail on unknown attribute/property")
                try:
                    return _read_energyml_xml_bytes_as_class(file, obj_type, False, False)
                except Exception as e:
                    logging.error(traceback.print_stack())
                    pass

        # Otherwise
        for obj_type_dev in get_energyml_class_in_related_dev_pkg(obj_type):
            try:
                logging.debug(f"Trying with class : {obj_type_dev}")
                obj = _read_energyml_xml_bytes_as_class(file, obj_type_dev)
                logging.debug(f" ==> succeed read with {obj_type_dev}")
                return obj
            except Exception:
                pass
        raise e


def read_energyml_xml_io(file: BytesIO, obj_class: Optional[type] = None) -> Any:
    if obj_class is not None:
        return _read_energyml_xml_bytes_as_class(file.getbuffer(), obj_class)
    else:
        return read_energyml_xml_bytes(file.getbuffer())


def read_energyml_xml_str(file_content: str) -> Any:
    encoding = get_xml_encoding(file_content)
    return read_energyml_xml_bytes(file_content.encode(encoding))


def read_energyml_xml_file(file_path: str) -> Any:
    xml_content_b = ""
    with open(file_path, "rb") as f:
        xml_content_b = f.read()
    return read_energyml_xml_bytes(xml_content_b)


def _read_energyml_json_bytes_as_class(file: bytes, json_version: JSON_VERSION, obj_class: type) -> Union[List, Any]:
    """
    Read a json file into energyml object. If json_version==JSON_VERSION.XSDATA the instance will be of type :param:`obj_class`.
    For json_version==JSON_VERSION.OSDU_OFFICIAL a list of read objects is returned
    :param file:
    :param json_version:
    :param obj_class:
    :return:
    """
    if json_version == JSON_VERSION.XSDATA:
        config = ParserConfig(
            # fail_on_unknown_properties=False,
            # fail_on_unknown_attributes=False,
            # process_xinclude=True,
        )
        parser = JsonParser(config=config)
        try:
            return parser.from_bytes(file, obj_class)
        except ParserError as e:
            logging.error(f"Failed to parse file {file} as class {obj_class}")
            raise e
    elif json_version == JSON_VERSION.OSDU_OFFICIAL:
        return read_json_dict(json.loads(file))


def read_energyml_json_bytes(
    file: Union[dict, bytes], json_version: JSON_VERSION, obj_type: Optional[type] = None
) -> Union[List, Any]:
    """
    Read a json file into energyml object. If json_version==JSON_VERSION.XSDATA the instance will be of type :param:`obj_class`.
    For json_version==JSON_VERSION.OSDU_OFFICIAL a list of read objects is returned
    :param file:
    :param json_version:
    :param obj_type: ignored if the json file is a list
    :return:
    """

    if isinstance(file, bytes):
        json_obj = json.loads(file)
    else:
        json_obj = file

    if not isinstance(json_obj, list):
        json_obj = [json_obj]
    else:
        obj_type = None

    result = []
    for obj in json_obj:
        try:
            if obj_type is None:
                obj_type = get_class_from_content_type(get_class_from_json_dict(obj))
            if json_version == JSON_VERSION.XSDATA:
                try:
                    result = result + _read_energyml_json_bytes_as_class(obj, obj_type)
                except xsdata.exceptions.ParserError as e:
                    logging.error(
                        f"Failed to read file with type {obj_type}: {get_energyml_class_in_related_dev_pkg(obj_type)}"
                    )
                    for obj_type_dev in get_energyml_class_in_related_dev_pkg(obj_type):
                        try:
                            logging.debug(f"Trying with class : {obj_type_dev}")
                            obj = _read_energyml_json_bytes_as_class(obj, obj_type_dev)
                            logging.debug(f" ==> succeed read with {obj_type_dev}")
                            result = result + obj
                        except Exception:
                            pass
                    raise e
            elif json_version == JSON_VERSION.OSDU_OFFICIAL:
                result = result + read_json_dict(obj)
        except Exception as e:
            logging.error(e)
            logging.error(obj)
            raise e
        obj_type = None

    return result


def read_energyml_json_io(
    file: BytesIO, json_version: JSON_VERSION = JSON_VERSION.OSDU_OFFICIAL, obj_class: Optional[type] = None
) -> Union[List, Any]:
    if obj_class is not None:
        return _read_energyml_json_bytes_as_class(file.getbuffer(), json_version, obj_class)
    else:
        return read_energyml_json_bytes(file.getbuffer(), json_version)


def read_energyml_json_str(
    file_content: str, json_version: JSON_VERSION = JSON_VERSION.OSDU_OFFICIAL
) -> Union[List, Any]:
    return read_energyml_json_bytes(file_content.encode("utf-8"), json_version)


def read_energyml_json_file(
    file_path: str, json_version: JSON_VERSION = JSON_VERSION.OSDU_OFFICIAL
) -> Union[List, Any]:
    json_content_b = ""
    with open(file_path, "rb") as f:
        json_content_b = f.read()
    return read_energyml_json_bytes(json_content_b, json_version)


#    _____           _       ___             __  _
#   / ___/___  _____(_)___ _/ (_)___  ____ _/ /_(_)___  ____
#   \__ \/ _ \/ ___/ / __ `/ / /_  / / __ `/ __/ / __ \/ __ \
#  ___/ /  __/ /  / / /_/ / / / / /_/ /_/ / /_/ / /_/ / / / /
# /____/\___/_/  /_/\__,_/_/_/ /___/\__,_/\__/_/\____/_/ /_/


def serialize_xml(obj) -> str:
    context = XmlContext(
        # element_name_generator=text.camel_case,
        # attribute_name_generator=text.kebab_case
    )
    serializer_config = SerializerConfig(indent="  ")
    serializer = XmlSerializer(context=context, config=serializer_config)
    return serializer.render(obj, ns_map=ENERGYML_NAMESPACES)


def serialize_json(obj, json_version: JSON_VERSION = JSON_VERSION.OSDU_OFFICIAL) -> str:
    if json_version == JSON_VERSION.XSDATA:
        context = XmlContext(
            # element_name_generator=text.camel_case,
            # attribute_name_generator=text.kebab_case
        )
        serializer_config = SerializerConfig(indent="  ")
        serializer = JsonSerializer(context=context, config=serializer_config)
        return serializer.render(obj)
    elif json_version == JSON_VERSION.OSDU_OFFICIAL:
        return json.dumps(to_json_dict(obj), indent=4, sort_keys=True)


def get_class_from_json_dict(o: Union[dict, bytes]) -> Optional[str]:
    """
    Searches for the attribute "$type"
    :param o:
    :return:
    """
    if isinstance(o, str) or isinstance(o, bytes):
        o = json.loads(o)
    for att in ["$type", "dataObjectType"]:
        if att in o:
            return o[att]
    return None


# RAW


def read_json_dict(obj: Any) -> List:
    """
    Reads a json dict valid with the OSDU standard.
    This means:
        - Any not "primitive" object (not str/number/bool ...) has a "$type" attribute set to its qualified type
        - None value are not given, except for mandatory attributes (depending on the energyml standard)
        - If an attribute is named 'value' (case-sensitive, this doesn't apply to 'Value'), the name of the attribute
            in the dict is "_"
        - "_data" attribute is given for DOR (not mandatory) and contains the json representation of the target object
    :param obj:
    :return: a list of read objects. This is a list due to the "_data" attribute
    """
    if "$type" in obj:
        sub_obj = []
        obj = _read_json_dict(obj, sub_obj)
        return [obj] + sub_obj
    else:
        raise UnknownTypeFromQualifiedType()


def _read_json_dict(obj_json: Any, sub_obj: List) -> Any:
    """
    Reads a json dict valid with the OSDU standard.
    This means:
        - Any not "primitive" object (not str/number/bool ...) has a "$type" attribute set to its qualified type
        - None value are not given, except for mandatory attributes (depending on the energyml standard)
        - If an attribute is named 'value' (case-sensitive, this doesn't apply to 'Value'), the name of the attribute
            in the dict is "_"
        - "_data" attribute is given for DOR (not mandatory) and contains the json representation of the target object
    :param obj_json:
    :param sub_obj: list of contextual external objects given inside the object that references them with a DOR
    :return: a list of read objects. This is a list due to the "_data" attribute
    """
    if isinstance(obj_json, dict) and "$type" in obj_json:
        qt = obj_json["$type"]

        obj_class = get_class_from_qualified_type(qt)
        if obj_class is None:
            raise UnknownTypeFromQualifiedType(qt + " " + json.dumps(obj_json))
        obj = obj_class()

        try:
            for att, val in obj_json.items():  # tous les autres attributs
                if att.lower() == "_data" and isinstance(val, dict):
                    for sub in read_json_dict(val):
                        sub_obj.append(sub)
                elif not att.startswith("$"):
                    if att == "_":
                        att = "value"
                    try:
                        matching = get_matching_class_attribute_name(obj, att)
                        if matching is not None:
                            setattr(
                                obj,
                                matching,
                                _read_json_dict(val, sub_obj),
                            )
                        else:
                            logging.error(f"No matching attribute for attribute {att} in {obj}")
                    except Exception as e:
                        logging.error(f"Error assign attribute value for attribute {att} in {obj}")
        except Exception as e:
            logging.error(
                f"Err on {att}",
                search_attribute_matching_name(
                    obj=obj,
                    name_rgx=att,
                    deep_search=False,
                    search_in_sub_obj=False,
                ),
                obj,
            )
            raise e
        return obj
    elif isinstance(obj_json, list):
        return [_read_json_dict(o, sub_obj) for o in obj_json]
    elif is_primitive(obj_json):
        # logging.debug(f"PRIM : {obj_json}")
        return obj_json
    else:
        raise NotParsableType(type(obj_json) + " " + obj_json)


def to_json_dict(obj: Any, obj_id_to_obj: Optional[Dict] = None) -> Any:
    """
    Transform an object to a dict valid with the OSDU standard
    :param obj:
    :param obj_id_to_obj:
    :return:
    """
    return to_json_dict_fn(
        obj,
        lambda _id: obj_id_to_obj[_id] if obj_id_to_obj is not None and _id in obj_id_to_obj else None,
    )


def to_json_dict_fn(obj: Any, f_identifier_to_obj: Callable) -> Any:
    """
    Transform an object to a dict valid with the OSDU standard
    :param obj:
    :param f_identifier_to_obj: A function that takes an object identifier see :func:`.introspection.get_obj_identifier`
            and returns the corresponding object
    :return:
    """
    assert f_identifier_to_obj is not None
    return _to_json_dict_fn(obj, f_identifier_to_obj, None)


def _fill_dict_with_attribs(
    res: Dict,
    obj: Any,
    f_identifier_to_obj: Optional[Callable] = None,
    _parent: Optional[Any] = None,
) -> None:

    for att_name, field in get_class_fields(obj).items():
        field_name = field.metadata["name"] if "name" in field.metadata else field.name
        if field_name == "value":
            field_name = "_"
        field_name = field_name[0].upper() + field_name[1:]
        mandatory = field.metadata["required"] if "required" in field.metadata else False
        value = getattr(obj, att_name)

        if "Any_element" in str(field_name):
            logging.debug(f"\t> {field_name}, {att_name} : {value}, {type(obj)}")

        if (value is not None or mandatory) and (not isinstance(value, list) or len(value) > 0):
            res[field_name] = _to_json_dict_fn(value, f_identifier_to_obj, obj)

            if _parent is not None and (field_name.lower() == "uuid" or field_name.lower() == "uid"):
                # adding referenced data
                ref_identifier = get_obj_identifier(obj)
                if f_identifier_to_obj is not None:
                    ref_value = f_identifier_to_obj(ref_identifier)
                    if ref_value is not None:
                        res["_data"] = to_json_dict_fn(ref_value, f_identifier_to_obj)
                    else:
                        logging.debug(f"NotFound : {ref_identifier}")


def _to_json_dict_fn(
    obj: Any,
    f_identifier_to_obj: Optional[Callable] = None,
    _parent: Optional[Any] = None,
) -> Any:
    """
    Transform an object to a dict valid with the OSDU standard
    :param obj:
    :param f_identifier_to_obj: A function that takes an object identifier see :func:`.introspection.get_obj_identifier`
            and returns the corresponding object
    :param _parent: None if :param:`obj` is the one given directly by the user, else the parent object of :param:`obj`
            in the original object given by the user
    :return: Any
    """
    if obj is None:
        return None
    elif is_enum(obj):
        return obj.value
        # return {
        #     "$type": get_qualified_type_from_class(obj),
        #     "_": obj.value
        # }
    elif is_primitive(obj):
        return obj
    elif isinstance(obj, xsdata.models.datatype.XmlDateTime):
        return str(obj)
    elif isinstance(obj, DerivedElement):
        res = {"$type": get_qualified_type_from_class(obj.value)}
        # _fill_dict_with_attribs(res, obj.value, f_identifier_to_obj, _parent)
        return res
    elif isinstance(obj, list):
        return [_to_json_dict_fn(o, f_identifier_to_obj, _parent) for o in obj]
    else:
        try:
            res = {"$type": get_qualified_type_from_class(obj)}
            _fill_dict_with_attribs(res, obj, f_identifier_to_obj, _parent)
            return res
        except Exception as e:
            logging.error(f"Except on qt: {obj} - {type(obj)}")
            raise e



# Copyright (c) 2023-2024 Geosiris.
# SPDX-License-Identifier: Apache-2.0
import re
from dataclasses import dataclass, field, Field
from enum import Enum
from typing import Any, List

from .epc import (
    get_obj_identifier,
    Epc,
)
from .introspection import (
    get_class_fields,
    get_object_attribute,
    search_attribute_matching_type_with_path,
    get_object_attribute_no_verif,
    get_object_attribute_rgx,
    get_matching_class_attribute_name,
    get_obj_uuid,
    get_obj_version,
    get_content_type_from_class,
    get_qualified_type_from_class,
    is_enum, get_object_uri,
)

class ErrorType(Enum):
    CRITICAL = "critical"
    DEBUG = "debug"
    INFO = "info"
    WARNING = "warning"


@dataclass
class ValidationError:

    msg: str = field(default="Validation error")

    error_type: ErrorType = field(default=ErrorType.INFO)

    def __str__(self):
        return f"[{str(self.error_type).upper()}] : {self.msg}"

    def toJson(self):
        return {
            "msg": self.msg,
            "error_type": self.error_type.value,
        }


@dataclass
class ValidationObjectError(ValidationError):

    target_obj: Any = field(default=None)

    attribute_dot_path: str = field(default=None)

    def __str__(self):
        return f"{ValidationError.__str__(self)}\n\t{get_obj_identifier(self.target_obj)} : '{self.attribute_dot_path}'"

    def toJson(self):
        return super().toJson() | {"target_obj": str(get_object_uri(self.target_obj)), "attribute_dot_path": self.attribute_dot_path}


@dataclass
class MandatoryError(ValidationObjectError):
    def __str__(self):
        return f"{ValidationError.__str__(self)}\n\tMandatory value is None for {get_obj_identifier(self.target_obj)} : '{self.attribute_dot_path}'"


@dataclass
class MissingEntityError(ValidationObjectError):
    missing_uuid: str = field(default=None)

    def __str__(self):
        return f"{ValidationError.__str__(self)}\n\tMissing entity in {get_obj_identifier(self.target_obj)} at path '{self.attribute_dot_path}'. Missing entity uuid: {self.missing_uuid}"


def validate_epc(epc: Epc) -> List[ValidationError]:
    """
    Verify if all :param:`epc`'s objects are valid.
    :param epc:
    :return:
    """
    errs = []
    for obj in epc.energyml_objects:
        errs = errs + patterns_validation(obj)

    errs = errs + dor_validation(epc.energyml_objects)

    return errs


def dor_validation(energyml_objects: List[Any]) -> List[ValidationError]:
    """
    Verification for DOR. An error is raised if DORs contains wrong information, or if a referenced object is unknown
    in the :param:`epc`.
    :param energyml_objects:
    :return:
    """
    errs = []

    dict_obj_identifier = {get_obj_identifier(obj): obj for obj in energyml_objects}
    dict_obj_uuid = {}
    for obj in energyml_objects:
        uuid = get_obj_uuid(obj)
        if uuid not in dict_obj_uuid:
            dict_obj_uuid[uuid] = []
        dict_obj_uuid[uuid].append(obj)

    # TODO: chercher dans les objets les AbstractObject (en Witsml des sous objet peuvent etre aussi references)

    for obj in energyml_objects:
        dor_list = search_attribute_matching_type_with_path(obj, "DataObjectReference")
        for dor_path, dor in dor_list:
            dor_target_id = get_obj_identifier(dor)
            if dor_target_id not in dict_obj_identifier:
                dor_uuid = get_obj_uuid(dor)
                dor_version = get_obj_version(dor)
                if dor_uuid not in dict_obj_uuid:
                    errs.append(
                        MissingEntityError(
                            error_type=ErrorType.CRITICAL,
                            target_obj=obj,
                            attribute_dot_path=dor_path,
                            missing_uuid=dor_uuid
                            # msg=f"[DOR ERR] has wrong information. Unkown object with uuid '{dor_uuid}'",
                        )
                    )
                else:
                    accessible_version = [get_obj_version(ref_obj) for ref_obj in dict_obj_uuid[dor_uuid]]
                    errs.append(
                        ValidationObjectError(
                            error_type=ErrorType.CRITICAL,
                            target_obj=obj,
                            attribute_dot_path=dor_path,
                            msg=f"[DOR ERR] has wrong information. Unkown object version '{dor_version}'. "
                            f"Version must be one of {accessible_version}",
                        )
                    )
            else:
                target = dict_obj_identifier[dor_target_id]
                target_title = get_object_attribute_rgx(target, "citation.title")
                target_content_type = get_content_type_from_class(target)
                target_qualified_type = get_qualified_type_from_class(target)

                dor_title = get_object_attribute_rgx(dor, "title")

                if dor_title != target_title:
                    errs.append(
                        ValidationObjectError(
                            error_type=ErrorType.CRITICAL,
                            target_obj=obj,
                            attribute_dot_path=dor_path,
                            msg=f"[DOR ERR] has wrong information. Title should be '{target_title}' and not '{dor_title}'",
                        )
                    )

                if get_matching_class_attribute_name(dor, "content_type") is not None:
                    dor_content_type = get_object_attribute_no_verif(dor, "content_type")
                    if dor_content_type != target_content_type:
                        errs.append(
                            ValidationObjectError(
                                error_type=ErrorType.CRITICAL,
                                target_obj=obj,
                                attribute_dot_path=dor_path,
                                msg=f"[DOR ERR] has wrong information. ContentType should be '{target_content_type}' and not '{dor_content_type}'",
                            )
                        )

                if get_matching_class_attribute_name(dor, "qualified_type") is not None:
                    dor_qualified_type = get_object_attribute_no_verif(dor, "qualified_type")
                    if dor_qualified_type != target_qualified_type:
                        errs.append(
                            ValidationObjectError(
                                error_type=ErrorType.CRITICAL,
                                target_obj=obj,
                                attribute_dot_path=dor_path,
                                msg=f"[DOR ERR] has wrong information. QualifiedType should be '{target_qualified_type}' and not '{dor_qualified_type}'",
                            )
                        )

    return errs


def patterns_validation(obj: Any) -> List[ValidationError]:
    """
    Verification on object values, using the patterns defined in the original energyml xsd files.
    :param obj:
    :return:
    """
    return _patterns_validation(obj, obj, "")


def _patterns_validation(obj: Any, root_obj: Any, current_attribute_dot_path: str = "") -> List[ValidationError]:
    """
    Verification on object values, using the patterns defined in the original energyml xsd files.
    :param obj:
    :param root_obj:
    :param current_attribute_dot_path:
    :return:
    """
    error_list = []

    if isinstance(obj, list):
        cpt = 0
        for val in obj:
            error_list = error_list + _patterns_validation(val, root_obj, f"{current_attribute_dot_path}.{cpt}")
            cpt = cpt + 1
    elif isinstance(obj, dict):
        for k, val in obj.items():
            error_list = error_list + _patterns_validation(val, root_obj, f"{current_attribute_dot_path}.{k}")
    else:
        # logging.debug(get_class_fields(obj))
        for att_name, att_field in get_class_fields(obj).items():
            # logging.debug(f"att_name : {att_field.metadata}")
            error_list = error_list + validate_attribute(
                get_object_attribute(obj, att_name, False),
                root_obj,
                att_field,
                f"{current_attribute_dot_path}.{att_name}",
            )

    return error_list


def validate_attribute(value: Any, root_obj: Any, att_field: Field, path: str) -> List[ValidationError]:
    errs = []

    if value is None:
        if att_field.metadata.get("required", False):
            errs.append(
                MandatoryError(
                    error_type=ErrorType.CRITICAL,
                    target_obj=root_obj,
                    attribute_dot_path=path,
                )
            )
    elif not is_enum(value):  # sometimes enums values fails the validation
        min_length = att_field.metadata.get("min_length", None)
        max_length = att_field.metadata.get("max_length", None)
        pattern = att_field.metadata.get("pattern", None)
        min_occurs = att_field.metadata.get("min_occurs", None)
        min_inclusive = att_field.metadata.get("min_inclusive", None)
        # white_space

        if max_length is not None:
            length = len(value)
            if length > max_length:
                errs.append(
                    ValidationObjectError(
                        error_type=ErrorType.CRITICAL,
                        target_obj=root_obj,
                        attribute_dot_path=path,
                        msg=f"Max length was {max_length} but found {length}",
                    )
                )

        if min_length is not None:
            length = len(value)
            if length < min_length:
                errs.append(
                    ValidationObjectError(
                        error_type=ErrorType.CRITICAL,
                        target_obj=root_obj,
                        attribute_dot_path=path,
                        msg=f"Max length was {min_length} but found {length}",
                    )
                )

        if min_occurs is not None:
            if isinstance(value, list) and min_occurs > len(value):
                errs.append(
                    ValidationObjectError(
                        error_type=ErrorType.CRITICAL,
                        target_obj=root_obj,
                        attribute_dot_path=path,
                        msg=f"Min occurs was {min_occurs} but found {len(value)}",
                    )
                )

        if min_inclusive is not None:
            potential_err = ValidationObjectError(
                error_type=ErrorType.CRITICAL,
                target_obj=root_obj,
                attribute_dot_path=path,
                msg=f"Min occurs was {min_inclusive} but found {value}",
            )
            if isinstance(value, list):
                for val in value:
                    if (isinstance(val, str) and len(val) > min_inclusive) or (
                        (isinstance(val, int) or isinstance(val, float)) and val > min_inclusive
                    ):
                        errs.append(potential_err)

        if pattern is not None:
            if not isinstance(value, list):
                testing_value_list = [value]
            else:
                testing_value_list = value

            for v in testing_value_list:
                if is_enum(v):
                    v = v.value
                if re.match(pattern, v) is None:
                    errs.append(
                        ValidationObjectError(
                            error_type=ErrorType.CRITICAL,
                            target_obj=root_obj,
                            attribute_dot_path=path,
                            msg=f"Pattern error. Value '{v}' was supposed to respect pattern '{pattern}'",
                        )
                    )

    return errs + _patterns_validation(
        obj=value,
        root_obj=root_obj,
        current_attribute_dot_path=path,
    )


def correct_dor(energyml_objects: List[Any]) -> None:
    """
    Fix DOR errors (missing object_version, wrong title, wrong content-type/qualified-type ...)
    :param energyml_objects:
    :return:
    """
    dict_obj_identifier = {get_obj_identifier(obj): obj for obj in energyml_objects}
    dict_obj_uuid = {}
    for obj in energyml_objects:
        uuid = get_obj_uuid(obj)
        if uuid not in dict_obj_uuid:
            dict_obj_uuid[uuid] = []
        dict_obj_uuid[uuid].append(obj)

    # TODO: chercher dans les objets les AbstractObject (en Witsml des sous objet peuvent etre aussi references)

    for obj in energyml_objects:
        dor_list = search_attribute_matching_type_with_path(obj, "DataObjectReference")
        for dor_path, dor in dor_list:
            dor_target_id = get_obj_identifier(dor)
            if dor_target_id in dict_obj_identifier:
                target = dict_obj_identifier[dor_target_id]
                target_title = get_object_attribute_rgx(target, "citation.title")
                target_content_type = get_content_type_from_class(target)
                target_qualified_type = get_qualified_type_from_class(target)

                dor_title = get_object_attribute_rgx(dor, "title")

                if dor_title != target_title:
                    dor.title = target_title

                if get_matching_class_attribute_name(dor, "content_type") is not None:
                    dor_content_type = get_object_attribute_no_verif(dor, "content_type")
                    if dor_content_type != target_content_type:
                        dor.content_type = target_content_type

                if get_matching_class_attribute_name(dor, "qualified_type") is not None:
                    dor_qualified_type = get_object_attribute_no_verif(dor, "qualified_type")
                    if dor_qualified_type != target_qualified_type:
                        dor.qualified_type = target_qualified_type



# Copyright (c) 2023-2024 Geosiris.
# SPDX-License-Identifier: Apache-2.0
"""
This module is a work in progress
"""  # pylint: disable=W0105

import logging
import os
import re
from dataclasses import dataclass
from io import BytesIO, TextIOWrapper, StringIO, BufferedReader
from typing import Optional, List, Tuple, Any, Union, TextIO, BinaryIO, Dict

import numpy as np

from .model import DatasetReader
from ..constants import EPCRelsRelationshipType, mime_type_to_file_extension
from ..exception import MissingExtraInstallation
from ..introspection import (
    search_attribute_matching_name_with_path,
    get_object_attribute,
    search_attribute_matching_name,
    get_obj_identifier,
    get_object_attribute_no_verif,
)

try:
    import h5py

    __H5PY_MODULE_EXISTS__ = True
except Exception as e:
    __H5PY_MODULE_EXISTS__ = False

try:
    import csv

    __CSV_MODULE_EXISTS__ = True
except Exception as e:
    __CSV_MODULE_EXISTS__ = False

try:
    import pandas as pd
    import pyarrow as pa
    import pyarrow.parquet as pq
    from pandas import DataFrame

    # import pyarrow.feather as feather
    __PARQUET_MODULE_EXISTS__ = True
except Exception as e:
    __PARQUET_MODULE_EXISTS__ = False

# HDF5
if __H5PY_MODULE_EXISTS__:

    @dataclass
    class HDF5FileReader(DatasetReader):
        def read_array(self, source: Union[BytesIO, str], path_in_external_file: str) -> Optional[List[Any]]:
            with h5py.File(source, "r") as f:
                d_group = f[path_in_external_file]
                return d_group[()].tolist()

        def get_array_dimension(self, source: Union[BytesIO, str], path_in_external_file: str) -> Optional[List[Any]]:
            with h5py.File(source, "r") as f:
                return list(f[path_in_external_file].shape)

        def extract_h5_datasets(
            self,
            input_h5: Union[BytesIO, str],
            output_h5: Union[BytesIO, str],
            h5_datasets_paths: List[str],
        ) -> None:
            """
            Copy all dataset from :param input_h5 matching with paths in :param h5_datasets_paths into the :param output
            :param input_h5:
            :param output_h5:
            :param h5_datasets_paths:
            :return:
            """
            if len(h5_datasets_paths) > 0:
                with h5py.File(output_h5, "w") as f_dest:
                    with h5py.File(input_h5, "r") as f_src:
                        for dataset in h5_datasets_paths:
                            f_dest.create_dataset(dataset, data=f_src[dataset])

    @dataclass
    class HDF5FileWriter:
        def write_array(
            self,
            target: Union[str, BytesIO, bytes],
            array: Union[list, np.array],
            path_in_external_file: str,
            dtype: Optional = None,
        ):
            if isinstance(array, list):
                array = np.asarray(array)
            print("writing array", target)
            with h5py.File(target, "a") as f:
                # print(array.dtype, h5py.string_dtype(), array.dtype == 'O')
                # print("\t", dtype or (h5py.string_dtype() if array.dtype == '0' else array.dtype))
                if array.dtype == "O":
                    array = np.asarray([s.encode() if isinstance(s, str) else s for s in array])
                    np.void(array)
                dset = f.create_dataset(path_in_external_file, array.shape, dtype or array.dtype)
                dset[()] = array

else:

    class HDF5FileReader:
        def read_array(self, source: Union[BytesIO, str], path_in_external_file: str) -> Optional[List[Any]]:
            raise MissingExtraInstallation(extra_name="hdf5")

        def get_array_dimension(self, source: Union[BytesIO, str], path_in_external_file: str) -> Optional[List[Any]]:
            raise MissingExtraInstallation(extra_name="hdf5")

        def extract_h5_datasets(
            self,
            input_h5: Union[BytesIO, str],
            output_h5: Union[BytesIO, str],
            h5_datasets_paths: List[str],
        ) -> None:
            raise MissingExtraInstallation(extra_name="hdf5")


# APACHE PARQUET
if __PARQUET_MODULE_EXISTS__:

    @dataclass
    class ParquetFileReader:
        def read_array(
            self, source: Union[BytesIO, str], path_in_external_file: Optional[str] = None
        ) -> Optional[List[Any]]:
            """
            :param source: the parquet file path or memory file
            :param path_in_external_file: the column name in the parquet file, if None, the entire table is returned
                (with a name for each column, not as a simple matrix)
            :return:
            """
            if isinstance(source, bytes):
                source = pa.BufferReader(source)
            array = pq.read_table(source)
            if path_in_external_file is not None:
                return array[path_in_external_file]
            else:
                return array
            # return pq.read_table(source).to_pandas()
            # return pq.read_table(source).to_pandas()[path_in_external_file]

        def get_array_dimension(self, source: Union[BytesIO, str], path_in_external_file: str) -> Optional[List[Any]]:
            if isinstance(source, bytes):
                source = pa.BufferReader(source)
            return pq.read_metadata(source)

    @dataclass
    class ParquetFileWriter:
        def write_array(
            self, target: Union[str, BytesIO, bytes], array: list, column_titles: Optional[List[str]] = None
        ) -> None:
            if (
                not isinstance(array[0], list)
                and not isinstance(array[0], np.ndarray)
                and not isinstance(array[0], pd.Series)
            ):
                # print(f"dtype : {type(array[0])}")
                array = [array]

            array_as_pd_df = pd.DataFrame({k: array[idx] for idx, k in enumerate(column_titles or range(len(array)))})

            pq.write_table(
                pa.Table.from_pandas(array_as_pd_df),
                target,
                version="2.6",
                compression="snappy",
            )

else:

    @dataclass
    class ParquetFileReader:
        def read_array(self, source: Union[BytesIO, str], path_in_external_file: Optional[str] = None) -> None:
            raise MissingExtraInstallation(extra_name="parquet")

        def get_array_dimension(self, source: Union[BytesIO, str], path_in_external_file: str) -> Optional[List[Any]]:
            raise MissingExtraInstallation(extra_name="parquet")

    @dataclass
    class ParquetFileWriter:
        def write_array(
            self, target: Union[BytesIO, str], array: list, column_titles: Optional[List[str]] = None
        ) -> None:
            raise MissingExtraInstallation(extra_name="parquet")


# DAT
@dataclass
class DATFileReader:
    def read_array(
        self,
        source: Union[BytesIO, TextIO, str],
        path_in_external_file: Optional[Union[int, str]] = None,
        delimiter: Optional[str] = ",",
        has_headers: bool = False,
        encoding: Optional[str] = "utf-8",
        **fmtparams,
    ) -> Optional[Dict]:
        """
        :param source: the dat file path or memory file
        :param path_in_external_file: the column name (or number) in the dat file, if None, the entire table is returned
            as it is in the file
        :param delimiter: the column delimiter
        :param has_headers: set it to True if the file contains column titles
        :param encoding:
        :return:
        """
        if isinstance(source, str):
            with open(source, "r", newline="") as datFile:
                return self.read_array(datFile, path_in_external_file, delimiter, has_headers, encoding, **fmtparams)
        else:
            comments = ""
            s_pos = 0
            c = source.readline()
            while c.startswith("#"):
                s_pos = source.tell()
                comments += c
                c = source.readline()

            source.seek(s_pos)

            logging.debug(comments)

            items = []

            if len(comments) > 0:
                _delim = re.search(r'Default\s+delimiter:\s*"(?P<delim>[^"])"', comments, re.IGNORECASE)
                logging.debug("delim", _delim, _delim.group("delim"))
                if _delim is not None:
                    _delim = _delim.group("delim")
                    logging.debug(_delim, "<==")
                    if len(_delim) > 0:
                        delimiter = _delim

                items = re.findall(
                    r"Item\s*:\s*(?P<itemName>[\w]+)\s+line\s+number\s*:\s*(?P<lnum>\d+)\s+delimiter\s+field\s+number\s*:\s*(?P<idx>\d+)",
                    comments,
                    re.IGNORECASE,
                )
                logging.debug("items", items)

                items = list(map(lambda it: (it[0], int(it[1]), int(it[2])), items))

                _cst = re.findall(
                    r"Item\s*:\s*(?P<itemName>[\w]+)\s+constant\s*:\s*(?P<value>\w+)", comments, re.IGNORECASE
                )
                logging.debug("cst", _cst)

                max_line_number = 0
                for _, n, _ in items:
                    if n > max_line_number:
                        max_line_number = n

                for i in range(max_line_number - 1):
                    source.readline()  # on skip les values des autres items, on ne garde que le tableau de valeurs
                logging.debug(max_line_number)
                logging.debug(items)
                # removing items not related to the columns titles items
                items = list(filter(lambda it: it[1] == max_line_number, items))
                logging.debug(items)

            if isinstance(source, BytesIO) or isinstance(source, BinaryIO) or isinstance(source, BufferedReader):
                source = TextIOWrapper(source, encoding=encoding)
            elif isinstance(source, bytes):
                source = StringIO(source.decode(encoding=encoding))

            if items is not None and len(items) > 0:
                return pd.read_csv(source, delimiter=delimiter, names=list(map(lambda it: it[0], items)), **fmtparams)
            else:
                array = csv.reader(source, delimiter=delimiter, **fmtparams)
                if path_in_external_file is not None and array is not None:
                    idx = int(path_in_external_file)
                    return [row[idx] for row in list(filter(lambda l: len(l) > 0, list(array)))]
                else:
                    return list(array)

    def read_array_as_panda_dict(
        self, source: Union[BytesIO, TextIO, str], delimiter: Optional[str] = ",", has_header: bool = True, **fmtparams
    ) -> Optional[Any]:
        if isinstance(source, str):
            with open(source, "r", newline="") as datFile:
                return self.read_array_as_panda_dict(datFile, delimiter, has_header=has_header, **fmtparams)
        else:
            return pd.read_csv(source, delimiter=delimiter, header=0 if has_header else None, **fmtparams)


# CSV
@dataclass
class CSVFileReader:
    def read_array(
        self,
        source: Union[BytesIO, TextIO, str],
        path_in_external_file: Optional[Union[int, str]] = None,
        delimiter: Optional[str] = ",",
        has_headers: bool = False,
        encoding: Optional[str] = "utf-8",
        **fmtparams,
    ) -> Optional[List[Any]]:
        """
        :param source: the csv file path or memory file
        :param path_in_external_file: the column name (or number) in the csv file, if None, the entire table is returned
            as it is in the file
        :param delimiter: the column delimiter
        :param has_headers: set it to True if the file contains column titles
        :return:
        """
        if isinstance(source, str):
            with open(source, "r", newline="") as csvFile:
                return self.read_array(csvFile, path_in_external_file, delimiter, has_headers, encoding, **fmtparams)
        else:
            if isinstance(source, BytesIO) or isinstance(source, BinaryIO) or isinstance(source, BufferedReader):
                source = TextIOWrapper(source, encoding=encoding)
            elif isinstance(source, bytes):
                source = StringIO(source.decode(encoding=encoding))

            if has_headers:
                if path_in_external_file is not None:
                    dictionary = csv.DictReader(source, delimiter=delimiter, **fmtparams)
                    if dictionary is not None:
                        return [row[path_in_external_file] for row in dictionary]
                else:
                    # array = pd.read_csv(source, delimiter=delimiter, **fmtparams)
                    # return [array.keys().tolist()] + array.values.tolist()
                    array = csv.reader(source, delimiter=delimiter, **fmtparams)
                    return list(array)
            else:
                array = csv.reader(source, delimiter=delimiter, **fmtparams)
                if path_in_external_file is not None and array is not None:
                    idx = int(path_in_external_file)
                    # for row in list(array):
                    #     print(len(row))
                    return [row[idx] for row in list(filter(lambda l: len(l) > 0, list(array)))]
                else:
                    return list(array)

    def read_array_as_panda_dict(
        self, source: Union[BytesIO, TextIO, str], delimiter: Optional[str] = ",", has_header: bool = True, **fmtparams
    ) -> Optional[Any]:
        if isinstance(source, str):
            with open(source, "r", newline="") as csvFile:
                return self.read_array_as_panda_dict(csvFile, delimiter, has_header=has_header, **fmtparams)
        else:
            return pd.read_csv(source, delimiter=delimiter, header=0 if has_header else None, **fmtparams)


@dataclass
class CSVFileWriter:
    def write_array(
        self,
        target: Union[BytesIO, TextIO, bytes, str],
        array: list,
        column_titles: Optional[List[str]] = None,
        delimiter: str = ",",
        **fmtparams,
    ) -> Optional[List[Any]]:
        if not isinstance(array[0], list):
            array = [array]

        if isinstance(target, str):
            with open(target, "w", newline="") as csvFile:
                return self.write_array(csvFile, array, column_titles, delimiter, **fmtparams)
        else:
            csvwriter = csv.writer(target, delimiter=delimiter, **fmtparams)
            csvwriter.writerows(array)


##############
def get_external_file_path_possibilities(
    value_in_xml: str, epc: Any, file_extension: Optional[str] = "h5"
) -> List[str]:
    """
    Maybe the path in the epc file objet was given as an absolute one : 'C:/my_file.h5'
    but if the epc has been moved (e.g. in 'D:/a_folder/') it will not work. Thus, the function
    energyml.utils.data.hdf.get_hdf5_path_from_external_path return the value from epc objet concatenate to the
    real epc folder path.
    With our example we will have : 'D:/a_folder/C:/my_file.h5'
    this function returns (following our example):
        [ 'C:/my_file.h5', 'D:/a_folder/my_file.h5', 'my_file.h5' ]
    :param value_in_xml:
    :param epc:
    :param file_extension: the external file extension (usually "h5" for "YOUR_HDF_FILE.h5")
    :return:
    """
    if epc is not None:
        epc_folder = epc.get_epc_file_folder()
        return get_external_file_path_possibilities_from_folder(file_raw_path=value_in_xml, folder_path=epc_folder) + [
            epc.epc_file_path[:-4] + f".{file_extension}",
        ]
    else:
        return get_external_file_path_possibilities_from_folder(file_raw_path=value_in_xml, folder_path=".")


def get_external_file_path_from_external_path(
    external_path_obj: Any,
    path_in_root: Optional[str] = None,
    root_obj: Optional[Any] = None,
    epc: Optional[Any] = None,
) -> Optional[List[str]]:
    """
    Return the hdf5 (or other type) file path (Searches for "uri" attribute or in :param:`epc` rels files).
    :param external_path_obj: can be an attribute of an ExternalDataArrayPart
    :param path_in_root:
    :param root_obj:
    :param epc: an EPC instance
    :return:
    """
    result = []
    if isinstance(external_path_obj, str):
        # external_path_obj is maybe an attribute of an ExternalDataArrayPart, now search upper in the object
        upper_path = path_in_root[: path_in_root.rindex(".")]
        result = get_external_file_path_from_external_path(
            external_path_obj=get_object_attribute(root_obj, upper_path),
            path_in_root=upper_path,
            root_obj=root_obj,
            epc=epc,
        )
    elif type(external_path_obj).__name__ == "ExternalDataArrayPart":
        # epc_folder = epc.get_epc_file_folder()
        external_file_uri = search_attribute_matching_name(external_path_obj, "uri")
        mimetype = next(iter(search_attribute_matching_name(external_path_obj, "MimeType")), None)
        if external_file_uri is not None and len(external_file_uri) > 0:
            result = get_external_file_path_possibilities(
                value_in_xml=external_file_uri[0], epc=epc, file_extension=mime_type_to_file_extension(mimetype) or "h5"
            )
            # result = f"{epc_folder}/{h5_uri[0]}"

    # epc_folder = epc.get_epc_file_folder()
    hdf_proxy_lst = search_attribute_matching_name(external_path_obj, "HdfProxy")
    ext_file_proxy_lst = search_attribute_matching_name(external_path_obj, "ExternalFileProxy")

    # resqml 2.0.1
    if hdf_proxy_lst is not None and len(hdf_proxy_lst) > 0:
        hdf_proxy = hdf_proxy_lst
        # logging.debug("h5Proxy", hdf_proxy)
        while isinstance(hdf_proxy, list):
            hdf_proxy = hdf_proxy[0]
        hdf_proxy_obj = epc.get_object_by_identifier(get_obj_identifier(hdf_proxy))
        try:
            logging.debug(f"hdf_proxy_obj : {hdf_proxy_obj} {hdf_proxy} : {hdf_proxy}")
        except:
            pass
        if hdf_proxy_obj is not None:
            for rel in epc.additional_rels.get(get_obj_identifier(hdf_proxy_obj), []):
                if rel.type_value == EPCRelsRelationshipType.EXTERNAL_RESOURCE.get_type():
                    result = get_external_file_path_possibilities(value_in_xml=rel.target, epc=epc)
                    # result = f"{epc_folder}/{rel.target}"

    # resqml 2.2dev3
    if ext_file_proxy_lst is not None and len(ext_file_proxy_lst) > 0:
        ext_file_proxy = ext_file_proxy_lst
        while isinstance(ext_file_proxy, list):
            ext_file_proxy = ext_file_proxy[0]
        ext_part_ref_obj = epc.get_object_by_identifier(
            get_obj_identifier(get_object_attribute_no_verif(ext_file_proxy, "epc_external_part_reference"))
        )
        result = get_external_file_path_possibilities(value_in_xml=ext_part_ref_obj.filename, epc=epc)
        # return f"{epc_folder}/{ext_part_ref_obj.filename}"

    # result += list(
    #     filter(
    #         lambda p: p.lower().endswith(".h5") or p.lower().endswith(".hdf5"),
    #         epc.external_files_path or [],
    #     )
    # )

    if epc is not None and len(result) == 0:
        result = [epc.epc_file_path[:-4] + ".h5"]

    try:
        logging.debug(f"{external_path_obj} {result} \n\t{hdf_proxy_lst}\n\t{ext_file_proxy_lst}")
    except:
        pass
    return result


def get_external_file_path_possibilities_from_folder(file_raw_path: str, folder_path: str) -> List[str]:
    external_path_respect = file_raw_path
    external_path_rematch = (
        f"{folder_path + '/' if folder_path is not None and len(folder_path) else ''}{os.path.basename(file_raw_path)}"
    )
    external_path_no_folder = f"{os.path.basename(file_raw_path)}"

    return [
        external_path_respect,
        external_path_rematch,
        external_path_no_folder,
    ]


def read_dataset(
    source: Union[BytesIO, str],
    path_in_external_file: Optional[str] = None,
    mimetype: Optional[str] = "application/x-hdf5",
) -> Any:
    mimetype = (mimetype or "").lower()
    file_reader = HDF5FileReader()  # default is hdf5
    if "parquet" in mimetype or (
        isinstance(source, str) and (source.lower().endswith(".parquet") or source.lower().endswith(".pqt"))
    ):
        file_reader = ParquetFileReader()
    elif "csv" in mimetype or (
        isinstance(source, str) and (source.lower().endswith(".csv") or source.lower().endswith(".dat"))
    ):
        file_reader = CSVFileReader()
    return file_reader.read_array(source, path_in_external_file)


def read_external_dataset_array(
    energyml_array: Any,
    root_obj: Optional[Any] = None,
    path_in_root: Optional[str] = None,
    additional_sources: Optional[List[Union[str, BytesIO, BufferedReader]]] = None,
    epc: Optional[any] = None,
):
    if additional_sources is None:
        additional_sources = []
    result_array = []

    for path_in_obj, path_in_external in get_path_in_external_with_path(energyml_array):
        succeed = False
        external_array_obj_path = ".".join(path_in_obj.split(".")[:-1])
        external_path_obj = get_object_attribute(obj=energyml_array, attr_dot_path=external_array_obj_path)
        mimetype = next(iter(search_attribute_matching_name(external_path_obj, "MimeType")), "application/x-hdf5")

        sources = additional_sources
        sources = sources + get_external_file_path_from_external_path(
            external_path_obj=external_path_obj, path_in_root=path_in_root, root_obj=root_obj, epc=epc
        )
        for s in sources:
            try:
                # TODO: take care of the "Counts" and "Starts" list in ExternalDataArrayPart to fill array correctly
                result_array = result_array + read_dataset(
                    source=s, path_in_external_file=path_in_external, mimetype=mimetype
                )
                succeed = True
                break  # stop after the first read success
            except MissingExtraInstallation as mei:
                raise mei
            except OSError as e:
                pass
        if not succeed:
            raise Exception(f"Failed to read external file. Paths tried : {sources}")

    return result_array


def get_path_in_external(obj) -> List[Any]:
    """
    See :func:`get_path_in_external_with_path`. Only the value is returned, not the dot path into the object
    :param obj:
    :return:
    """
    return [val for path, val in get_path_in_external_with_path(obj=obj)]


def get_path_in_external_with_path(obj: any) -> List[Tuple[str, Any]]:
    """
    See :func:`search_attribute_matching_name_with_path`. Search an attribute with type matching regex
    "(PathInHdfFile|PathInExternalFile)".

    :param obj:
    :return: [ (Dot_Path_In_Obj, value), ...]
    """
    return search_attribute_matching_name_with_path(obj, "(PathInHdfFile|PathInExternalFile)")



# Copyright (c) 2023-2024 Geosiris.
# SPDX-License-Identifier: Apache-2.0
"""
The data module.

Contains functions to help the read of specific entities like Grid2DRepresentation, TriangulatedSetRepresentation etc.
It also contains functions to export data into OFF/OBJ format.
"""



# Copyright (c) 2023-2024 Geosiris.
# SPDX-License-Identifier: Apache-2.0
import importlib
import inspect
import logging
import pkgutil
from typing import Union, Any, Dict

from .constants import *


def get_related_energyml_modules_name(cls: Union[type, Any]) -> List[str]:
    """
    Return the list of all energyml modules related to another one.
    For example resqml 2.0.1 is related to common 2.0
    :param cls:
    :return:
    """
    if isinstance(cls, type):
        for related in RELATED_MODULES:
            if cls.__module__ in related:
                return related
    else:
        return get_related_energyml_modules_name(type(cls))
    return []


def dict_energyml_modules() -> Dict:
    """
    List all accessible energyml python modules
    :return:
    """
    modules = {}

    energyml_module = importlib.import_module("energyml")
    # logging.debug("> energyml")

    for mod in pkgutil.iter_modules(energyml_module.__path__):
        # logging.debug(f"{mod.name}")
        if mod.name in ENERGYML_MODULES_NAMES:
            energyml_sub_module = importlib.import_module(f"energyml.{mod.name}")
            if mod.name not in modules:
                modules[mod.name] = []
            for sub_mod in pkgutil.iter_modules(energyml_sub_module.__path__):
                modules[mod.name].append(sub_mod.name)
                # modules[mod.name].append(re.sub(r"^\D*(?P<number>\d+(.\d+)*$)",
                # r"\g<number>", sub_mod.name).replace("_", "."))
    return modules


def list_energyml_modules() -> List:
    try:
        energyml_module = importlib.import_module("energyml")
        modules = []
        for obj in pkgutil.iter_modules(energyml_module.__path__):
            # logging.debug(f"{obj.name}")
            if obj.name in ENERGYML_MODULES_NAMES:
                modules.append(obj.name)
        return modules
    except ModuleNotFoundError:
        return []


def list_classes(module_path: str) -> List:
    """
    List all accessible classes from a specific module
    :param module_path:
    :return:
    """
    try:
        module = importlib.import_module(module_path)
        class_list = []
        for _, obj in inspect.getmembers(module):
            if inspect.isclass(obj):
                class_list.append(obj)
        return class_list
    except ModuleNotFoundError:
        logging.error(f"Err : module {module_path} not found")
        return []


def get_sub_classes(cls: type) -> List[type]:
    """
    Return all classes that extends the class :param:`cls`.
    :param cls:
    :return:
    """
    sub_classes = []
    for related in get_related_energyml_modules_name(cls):
        try:
            module = importlib.import_module(related)
            for _, obj in inspect.getmembers(module):
                if inspect.isclass(obj) and cls in obj.__bases__:
                    sub_classes.append(obj)
                    sub_classes = sub_classes + get_sub_classes(obj)
        except ModuleNotFoundError:
            pass
    return list(dict.fromkeys(sub_classes))


def get_classes_matching_name(
    cls: type,
    name_rgx: str,
    re_flags=re.IGNORECASE,
) -> List[type]:
    """
    Search a class matching the regex @re_flags. The search is the energyml packages related to the objet type @cls.
    :param cls:
    :param name_rgx:
    :param re_flags:
    :return:
    """
    match_classes = []
    for related in get_related_energyml_modules_name(cls):
        try:
            module = importlib.import_module(related)
            for _, obj in inspect.getmembers(module):
                if inspect.isclass(obj) and re.match(name_rgx, obj.__name__, re_flags):
                    match_classes.append(obj)
        except ModuleNotFoundError:
            pass
    return list(dict.fromkeys(match_classes))


def get_all_energyml_classes() -> dict:
    result = {}
    for mod_name, versions in dict_energyml_modules().items():
        for version in versions:
            result = result | get_all_classes(mod_name, version)
    return result


def get_all_classes(module_name: str, version: str) -> dict:
    result = {}
    pkg_path = f"energyml.{module_name}.{version}"
    package = importlib.import_module(pkg_path)
    for _, modname, _ in pkgutil.walk_packages(
        path=getattr(package, "__path__"),
        prefix=package.__name__ + ".",
        onerror=lambda x: None,
    ):
        result[pkg_path] = []
        for classFound in list_classes(modname):
            try:
                result[pkg_path].append(classFound)
            except Exception:
                pass

    return result


def get_class_pkg(cls):
    try:
        p = re.compile(RGX_ENERGYML_MODULE_NAME)
        match = p.search(cls.__module__)
        return match.group("pkg")
    except AttributeError as e:
        logging.error(f"Exception to get class package for '{cls}'")
        raise e


def reshape_version(version: str, nb_digit: int) -> str:
    """
    Reshape a project version to have only specific number of digits. If 0 < nbDigit < 4 then the reshape is done,
    else, the original version is returned.
    Example : reshapeVersion("v2.0.1", 2) ==> "2.0" and reshapeVersion("version2.0.1.3.2.5", 4) ==> "version2.0.1.3.2.5"
    """
    p = re.compile(RGX_PROJECT_VERSION)
    match = p.search(version)
    if match is not None:
        n0 = match.group("n0")
        n1 = match.group("n1")
        n2 = match.group("n2")
        if nb_digit == 1:
            return n0
        elif nb_digit == 2:
            return n0 + ("." + n1 if n1 is not None else "")
        elif nb_digit == 3:
            return n0 + ("." + n1 + ("." + n2 if n2 is not None else "") if n1 is not None else "")

    return version


def get_class_pkg_version(cls, print_dev_version: bool = True, nb_max_version_digits: int = 2):
    p = re.compile(RGX_ENERGYML_MODULE_NAME)
    class_module = None
    if isinstance(cls, type):
        class_module = cls.__module__
    elif isinstance(cls, str):
        class_module = cls
    else:
        class_module = type(cls).__module__

    match = p.search(class_module)
    return reshape_version(match.group("versionNumber"), nb_max_version_digits) + (
        "dev" + match.group("versionDev") if match.group("versionDev") is not None and print_dev_version else ""
    )


# ProtocolDict = DefaultDict[str, MessageDict]
# def get_all__classes() -> ProtocolDict:
#     protocolDict: ProtocolDict = defaultdict(
#         lambda: defaultdict(type(ETPModel))
#     )
#     package = energyml
#     for _, modname, _ in pkgutil.walk_packages(
#         path=getattr(package, "__path__"),
#         prefix=package.__name__ + ".",
#         onerror=lambda x: None,
#     ):
#         for classFound in list_classes(modname):
#             try:
#                 schem = json.loads(avro_schema(classFound))
#                 protocolId = schem["protocol"]
#                 messageType = schem["messageType"]
#                 protocolDict[protocolId][messageType] = classFound
#             except Exception:
#                 pass
#     return protocolDict



# Copyright (c) 2023-2024 Geosiris.
# SPDX-License-Identifier: Apache-2.0
from dataclasses import dataclass
from io import BytesIO
from typing import Optional, List, Any, Union


@dataclass
class DatasetReader:
    def read_array(self, source: Union[BytesIO, str], path_in_external_file: str) -> Optional[List[Any]]:
        return None

    def get_array_dimension(self, source: Union[BytesIO, str], path_in_external_file: str) -> Optional[List[Any]]:
        return None


@dataclass
class ETPReader(DatasetReader):
    def read_array(self, obj_uri: str, path_in_external_file: str) -> Optional[List[Any]]:
        return None

    def get_array_dimension(self, source: str, path_in_external_file: str) -> Optional[List[Any]]:
        return None



# Copyright (c) 2023-2024 Geosiris.
# SPDX-License-Identifier: Apache-2.0
import inspect
import json
import logging
import os
import re
import sys
from dataclasses import dataclass, field
from enum import Enum
from io import BytesIO
from typing import List, Optional, Any, Callable, Dict, Union, Tuple


from .helper import (
    read_array,
    read_grid2d_patch,
    EnergymlWorkspace,
    get_crs_obj,
    get_crs_origin_offset,
    is_z_reversed,
)
from ..epc import Epc, get_obj_identifier, gen_energyml_object_path
from ..exception import ObjectNotFoundNotError
from ..introspection import (
    search_attribute_matching_name,
    search_attribute_matching_name_with_path,
    snake_case,
    get_object_attribute,
)

_FILE_HEADER: bytes = b"# file exported by energyml-utils python module (Geosiris)\n"

Point = list[float]


class MeshFileFormat(Enum):
    OFF = "off"
    OBJ = "obj"
    GEOJSON = "geojson"


class GeoJsonGeometryType(Enum):
    """GeoJson type enum"""

    Point = "Point"
    MultiPoint = "MultiPoint"
    LineString = "LineString"
    MultiLineString = "MultiLineString"
    Polygon = "Polygon"
    MultiPolygon = "MultiPolygon"


def energyml_type_to_geojson_type(energyml_type: str):
    if "PolylineSet" in energyml_type:
        return GeoJsonGeometryType.MultiLineString
    elif "Polyline" in energyml_type:
        return GeoJsonGeometryType.LineString
    elif "PointSet" in energyml_type:
        return GeoJsonGeometryType.MultiPoint
    elif "Point" in energyml_type:
        return GeoJsonGeometryType.Point
    elif "TriangulatedSet" in energyml_type:
        return GeoJsonGeometryType.MultiPolygon
    elif "Triangulated" in energyml_type:
        return GeoJsonGeometryType.Polygon
    elif "Grid2" in energyml_type:
        return GeoJsonGeometryType.MultiPolygon
    return GeoJsonGeometryType.Point


@dataclass
class AbstractMesh:
    energyml_object: Any = field(default=None)

    crs_object: Any = field(default=None)

    point_list: List[Point] = field(
        default_factory=list,
    )

    identifier: str = field(
        default=None,
    )

    def get_nb_edges(self) -> int:
        return 0

    def get_nb_faces(self) -> int:
        return 0

    def get_indices(self) -> List[List[int]]:
        return []


@dataclass
class PointSetMesh(AbstractMesh):
    pass


@dataclass
class PolylineSetMesh(AbstractMesh):
    line_indices: List[List[int]] = field(
        default_factory=list,
    )

    def get_nb_edges(self) -> int:
        return sum(list(map(lambda li: len(li) - 1, self.line_indices)))

    def get_nb_faces(self) -> int:
        return 0

    def get_indices(self) -> List[List[int]]:
        return self.line_indices


@dataclass
class SurfaceMesh(AbstractMesh):
    faces_indices: List[List[int]] = field(
        default_factory=list,
    )

    def get_nb_edges(self) -> int:
        return sum(list(map(lambda li: len(li) - 1, self.faces_indices)))

    def get_nb_faces(self) -> int:
        return len(self.faces_indices)

    def get_indices(self) -> List[List[int]]:
        return self.faces_indices


def crs_displacement(points: List[Point], crs_obj: Any) -> Tuple[List[Point], Point]:
    """
    Transform a point list with CRS information (XYZ offset and ZIncreasingDownward)
    :param points: in/out : the list is directly modified
    :param crs_obj:
    :return: The translated points and the crs offset vector.
    """
    crs_point_offset = get_crs_origin_offset(crs_obj=crs_obj)
    zincreasing_downward = is_z_reversed(crs_obj)

    if crs_point_offset != [0, 0, 0]:
        for p in points:
            for xyz in range(len(p)):
                p[xyz] = p[xyz] + crs_point_offset[xyz]
            if zincreasing_downward and len(p) >= 3:
                p[2] = -p[2]

    return points, crs_point_offset


def get_mesh_reader_function(mesh_type_name: str) -> Optional[Callable]:
    """
    Returns the name of the potential appropriate function to read an object with type is named mesh_type_name
    :param mesh_type_name: the initial type name
    :return:
    """
    for name, obj in inspect.getmembers(sys.modules[__name__]):
        if name == f"read_{snake_case(mesh_type_name)}":
            return obj
    return None


def _mesh_name_mapping(array_type_name: str) -> str:
    """
    Transform the type name to match existing reader function
    :param array_type_name:
    :return:
    """
    array_type_name = array_type_name.replace("3D", "3d").replace("2D", "2d")
    array_type_name = re.sub(r"^[Oo]bj([A-Z])", r"\1", array_type_name)
    array_type_name = re.sub(r"(Polyline|Point)Set", r"\1", array_type_name)
    return array_type_name


def read_mesh_object(
    energyml_object: Any,
    workspace: Optional[EnergymlWorkspace] = None,
    use_crs_displacement: bool = False,
    sub_indices: List[int] = None,
) -> List[AbstractMesh]:
    """
    Read and "meshable" object. If :param:`energyml_object` is not supported, an exception will be raised.
    :param energyml_object:
    :param workspace:
    :param use_crs_displacement: If true the :py:function:`crs_displacement <energyml.utils.mesh.crs_displacement>`
    is used to translate the data with the CRS offsets
    :return:
    """
    if isinstance(energyml_object, list):
        return energyml_object
    array_type_name = _mesh_name_mapping(type(energyml_object).__name__)

    reader_func = get_mesh_reader_function(array_type_name)
    if reader_func is not None:
        surfaces: List[AbstractMesh] = reader_func(
            energyml_object=energyml_object, workspace=workspace, sub_indices=sub_indices
        )
        if use_crs_displacement:
            for s in surfaces:
                crs_displacement(s.point_list, s.crs_object)
        return surfaces
    else:
        logging.error(f"Type {array_type_name} is not supported: function read_{snake_case(array_type_name)} not found")
        raise Exception(
            f"Type {array_type_name} is not supported\n\t{energyml_object}: \n\tfunction read_{snake_case(array_type_name)} not found"
        )


def read_point_representation(
    energyml_object: Any, workspace: EnergymlWorkspace, sub_indices: List[int] = None
) -> List[PointSetMesh]:
    # pt_geoms = search_attribute_matching_type(point_set, "AbstractGeometry")

    meshes = []

    patch_idx = 0
    total_size = 0
    for (
        points_path_in_obj,
        points_obj,
    ) in search_attribute_matching_name_with_path(
        energyml_object, r"NodePatch.[\d]+.Geometry.Points"
    ) + search_attribute_matching_name_with_path(  # resqml 2.0.1
        energyml_object, r"NodePatchGeometry.[\d]+.Points"
    ):  # resqml 2.2
        points = read_array(
            energyml_array=points_obj,
            root_obj=energyml_object,
            path_in_root=points_path_in_obj,
            workspace=workspace,
        )

        crs = None
        try:
            crs = get_crs_obj(
                context_obj=points_obj,
                path_in_root=points_path_in_obj,
                root_obj=energyml_object,
                workspace=workspace,
            )
        except ObjectNotFoundNotError as e:
            logging.error(e)
            pass

        if sub_indices is not None and len(sub_indices) > 0:
            new_points = []
            for idx in sub_indices:
                t_idx = idx - total_size
                if 0 <= t_idx < len(points):
                    new_points.append(points[t_idx])
            total_size = total_size + len(points)
            points = new_points
        # else:
        #     total_size = total_size + len(points)

        if points is not None:
            meshes.append(
                PointSetMesh(
                    identifier=f"Patch num {patch_idx}",
                    energyml_object=energyml_object,
                    crs_object=crs,
                    point_list=points,
                )
            )

        patch_idx = patch_idx + 1

    return meshes


def read_polyline_representation(
    energyml_object: Any, workspace: EnergymlWorkspace, sub_indices: List[int] = None
) -> List[PolylineSetMesh]:
    # pt_geoms = search_attribute_matching_type(point_set, "AbstractGeometry")

    meshes = []

    patch_idx = 0
    total_size = 0
    for patch_path_in_obj, patch in search_attribute_matching_name_with_path(
        energyml_object, "NodePatch"
    ) + search_attribute_matching_name_with_path(energyml_object, r"LinePatch.[\d]+"):
        points_path, points_obj = search_attribute_matching_name_with_path(patch, "Geometry.Points")[0]
        points = read_array(
            energyml_array=points_obj,
            root_obj=energyml_object,
            path_in_root=patch_path_in_obj + "." + points_path,
            workspace=workspace,
        )

        crs = None
        try:
            crs = get_crs_obj(
                context_obj=points_obj,
                path_in_root=patch_path_in_obj + "." + points_path,
                root_obj=energyml_object,
                workspace=workspace,
            )
        except ObjectNotFoundNotError as e:
            logging.error(e)

        close_poly = None
        try:
            (
                close_poly_path,
                close_poly_obj,
            ) = search_attribute_matching_name_with_path(
                patch, "ClosedPolylines"
            )[0]
            close_poly = read_array(
                energyml_array=close_poly_obj,
                root_obj=energyml_object,
                path_in_root=patch_path_in_obj + "." + close_poly_path,
                workspace=workspace,
            )
        except IndexError:
            pass

        point_indices = []
        try:
            (
                node_count_per_poly_path_in_obj,
                node_count_per_poly,
            ) = search_attribute_matching_name_with_path(
                patch, "NodeCountPerPolyline"
            )[0]
            node_counts_list = read_array(
                energyml_array=node_count_per_poly,
                root_obj=energyml_object,
                path_in_root=patch_path_in_obj + node_count_per_poly_path_in_obj,
                workspace=workspace,
            )
            idx = 0
            poly_idx = 0
            for nb_node in node_counts_list:
                point_indices.append([x for x in range(idx, idx + nb_node)])
                if close_poly is not None and len(close_poly) > poly_idx and close_poly[poly_idx]:
                    point_indices[len(point_indices) - 1].append(idx)
                idx = idx + nb_node
                poly_idx = poly_idx + 1
        except IndexError:
            # No NodeCountPerPolyline for Polyline but only in PolylineSet
            pass

        if point_indices is None or len(point_indices) == 0:
            # No indices ==> all point in the polyline
            point_indices = [list(range(len(points)))]

        if sub_indices is not None and len(sub_indices) > 0:
            new_indices = []
            for idx in sub_indices:
                t_idx = idx - total_size
                if 0 <= t_idx < len(point_indices):
                    new_indices.append(point_indices[t_idx])
            total_size = total_size + len(point_indices)
            point_indices = new_indices
        else:
            total_size = total_size + len(point_indices)

        if len(points) > 0:
            meshes.append(
                PolylineSetMesh(
                    identifier=f"{get_obj_identifier(energyml_object)}_patch{patch_idx}",
                    energyml_object=energyml_object,
                    crs_object=crs,
                    point_list=points,
                    line_indices=point_indices,
                )
            )

        patch_idx = patch_idx + 1

    return meshes


def gen_surface_grid_geometry(
    energyml_object: Any,
    patch: Any,
    patch_path: Any,
    workspace: Optional[EnergymlWorkspace] = None,
    keep_holes=False,
    sub_indices: List[int] = None,
    offset: int = 0,
):
    points = read_grid2d_patch(
        patch=patch,
        grid2d=energyml_object,
        path_in_root=patch_path,
        workspace=workspace,
    )

    fa_count = search_attribute_matching_name(patch, "FastestAxisCount")
    if fa_count is None:
        fa_count = search_attribute_matching_name(energyml_object, "FastestAxisCount")

    sa_count = search_attribute_matching_name(patch, "SlowestAxisCount")
    if sa_count is None:
        sa_count = search_attribute_matching_name(energyml_object, "SlowestAxisCount")

    fa_count = fa_count[0]
    sa_count = sa_count[0]

    # logging.debug(f"sa_count {sa_count} fa_count {fa_count}")

    points_no_nan = []

    indice_to_final_indice = {}
    if keep_holes:
        for i in range(len(points)):
            p = points[i]
            if p[2] != p[2]:  # a NaN
                points[i][2] = 0
    else:
        for i in range(len(points)):
            p = points[i]
            if p[2] == p[2]:  # not a NaN
                indice_to_final_indice[i] = len(points_no_nan)
                points_no_nan.append(p)
    indices = []

    while sa_count * fa_count > len(points):
        sa_count = sa_count - 1
        fa_count = fa_count - 1

    while sa_count * fa_count < len(points):
        sa_count = sa_count + 1
        fa_count = fa_count + 1

    # logging.debug(f"sa_count {sa_count} fa_count {fa_count} : {sa_count*fa_count} - {len(points)} ")

    for sa in range(sa_count - 1):
        for fa in range(fa_count - 1):
            line = sa * fa_count
            # if sa+1 == int(sa_count / 2) and fa == int(fa_count / 2):
            #     logging.debug(
            #         "\n\t", (line + fa), " : ", (line + fa) in indice_to_final_indice,
            #         "\n\t", (line + fa + 1), " : ", (line + fa + 1) in indice_to_final_indice,
            #         "\n\t", (line + fa_count + fa + 1), " : ", (line + fa_count + fa + 1) in indice_to_final_indice,
            #         "\n\t", (line + fa_count + fa), " : ", (line + fa_count + fa) in indice_to_final_indice,
            #     )
            if keep_holes:
                indices.append(
                    [
                        line + fa,
                        line + fa + 1,
                        line + fa_count + fa + 1,
                        line + fa_count + fa,
                    ]
                )
            elif (
                (line + fa) in indice_to_final_indice
                and (line + fa + 1) in indice_to_final_indice
                and (line + fa_count + fa + 1) in indice_to_final_indice
                and (line + fa_count + fa) in indice_to_final_indice
            ):
                indices.append(
                    [
                        indice_to_final_indice[line + fa],
                        indice_to_final_indice[line + fa + 1],
                        indice_to_final_indice[line + fa_count + fa + 1],
                        indice_to_final_indice[line + fa_count + fa],
                    ]
                )
    if sub_indices is not None and len(sub_indices) > 0:
        new_indices = []
        for idx in sub_indices:
            t_idx = idx - offset
            if 0 <= t_idx < len(indices):
                new_indices.append(indices[t_idx])
        indices = new_indices
    # logging.debug(indices)

    return points if keep_holes else points_no_nan, indices


def read_grid2d_representation(
    energyml_object: Any, workspace: Optional[EnergymlWorkspace] = None, keep_holes=False, sub_indices: List[int] = None
) -> List[SurfaceMesh]:
    # h5_reader = HDF5FileReader()
    meshes = []

    if sub_indices is not None:
        sub_indices = list(sorted(sub_indices))

    patch_idx = 0
    total_size = 0

    # Resqml 201
    for patch_path, patch in search_attribute_matching_name_with_path(energyml_object, "Grid2dPatch"):
        crs = None
        try:
            crs = get_crs_obj(
                context_obj=patch,
                path_in_root=patch_path,
                root_obj=energyml_object,
                workspace=workspace,
            )
        except ObjectNotFoundNotError as e:
            pass

        points, indices = gen_surface_grid_geometry(
            energyml_object=energyml_object,
            patch=patch,
            patch_path=patch_path,
            workspace=workspace,
            keep_holes=keep_holes,
            sub_indices=sub_indices,
            offset=total_size,
        )

        total_size = total_size + len(indices)

        meshes.append(
            SurfaceMesh(
                identifier=f"{get_obj_identifier(energyml_object)}_patch{patch_idx}",
                energyml_object=energyml_object,
                crs_object=crs,
                point_list=points,
                faces_indices=indices,
            )
        )
        patch_idx = patch_idx + 1

    # Resqml 22
    if hasattr(energyml_object, "geometry"):
        crs = None
        try:
            crs = get_crs_obj(
                context_obj=energyml_object,
                path_in_root=".",
                root_obj=energyml_object,
                workspace=workspace,
            )
        except ObjectNotFoundNotError as e:
            logging.error(e)
        # geometry = energyml_object.geometry
        # points = read_grid2d_patch(
        #     patch=energyml_object,
        #     grid2d=energyml_object,
        #     path_in_root="",
        #     workspace=workspace,
        # )
        points, indices = gen_surface_grid_geometry(
            energyml_object=energyml_object,
            patch=energyml_object,
            patch_path="",
            workspace=workspace,
            keep_holes=keep_holes,
            sub_indices=sub_indices,
            offset=total_size,
        )
        meshes.append(
            SurfaceMesh(
                identifier=f"{get_obj_identifier(energyml_object)}_patch{patch_idx}",
                energyml_object=energyml_object,
                crs_object=crs,
                point_list=points,
                faces_indices=indices,
            )
        )

    return meshes


def read_triangulated_set_representation(
    energyml_object: Any,
    workspace: EnergymlWorkspace,
    sub_indices: List[int] = None,
) -> List[SurfaceMesh]:
    meshes = []

    point_offset = 0
    patch_idx = 0
    total_size = 0
    for patch_path, patch in search_attribute_matching_name_with_path(
        energyml_object,
        "\\.*Patch.\\d+",
        deep_search=False,
        search_in_sub_obj=False,
    ):
        crs = None
        try:
            crs = get_crs_obj(
                context_obj=patch,
                path_in_root=patch_path,
                root_obj=energyml_object,
                workspace=workspace,
            )
        except ObjectNotFoundNotError as e:
            pass

        point_list: List[Point] = []
        for point_path, point_obj in search_attribute_matching_name_with_path(patch, "Geometry.Points"):
            point_list = point_list + read_array(
                energyml_array=point_obj,
                root_obj=energyml_object,
                path_in_root=patch_path + "." + point_path,
                workspace=workspace,
            )

        triangles_list: List[List[int]] = []
        for (
            triangles_path,
            triangles_obj,
        ) in search_attribute_matching_name_with_path(patch, "Triangles"):
            triangles_list = triangles_list + read_array(
                energyml_array=triangles_obj,
                root_obj=energyml_object,
                path_in_root=patch_path + "." + triangles_path,
                workspace=workspace,
            )
        triangles_list = list(map(lambda tr: [ti - point_offset for ti in tr], triangles_list))
        if sub_indices is not None and len(sub_indices) > 0:
            new_triangles_list = []
            for idx in sub_indices:
                t_idx = idx - total_size
                if 0 <= t_idx < len(triangles_list):
                    new_triangles_list.append(triangles_list[t_idx])
            total_size = total_size + len(triangles_list)
            triangles_list = new_triangles_list
        else:
            total_size = total_size + len(triangles_list)
        meshes.append(
            SurfaceMesh(
                identifier=f"{get_obj_identifier(energyml_object)}_patch{patch_idx}",
                energyml_object=energyml_object,
                crs_object=crs,
                point_list=point_list,
                faces_indices=triangles_list,
            )
        )
        point_offset = point_offset + len(point_list)
        patch_idx += 1

    return meshes


def read_sub_representation(
    energyml_object: Any,
    workspace: EnergymlWorkspace,
    sub_indices: List[int] = None,
) -> List[AbstractMesh]:
    supporting_rep_dor = search_attribute_matching_name(
        obj=energyml_object, name_rgx=r"(SupportingRepresentation|RepresentedObject)"
    )[0]
    supporting_rep_identifier = get_obj_identifier(supporting_rep_dor)
    supporting_rep = workspace.get_object_by_identifier(supporting_rep_identifier)

    total_size = 0
    all_indices = []
    for patch_path, patch_indices in search_attribute_matching_name_with_path(
        obj=energyml_object,
        name_rgx="SubRepresentationPatch.\\d+.ElementIndices.\\d+.Indices",
        deep_search=False,
        search_in_sub_obj=False,
    ) + search_attribute_matching_name_with_path(
        obj=energyml_object,
        name_rgx="SubRepresentationPatch.\\d+.Indices",
        deep_search=False,
        search_in_sub_obj=False,
    ):
        array = read_array(
            energyml_array=patch_indices,
            root_obj=energyml_object,
            path_in_root=patch_path,
            workspace=workspace,
            sub_indices=sub_indices,
        )

        if sub_indices is not None and len(sub_indices) > 0:
            new_array = []
            for idx in sub_indices:
                t_idx = idx - total_size
                if 0 <= t_idx < len(array):
                    new_array.append(array[t_idx])
            total_size = total_size + len(array)
            array = new_array
        else:
            total_size = total_size + len(array)

        all_indices = all_indices + array
    meshes = read_mesh_object(
        energyml_object=supporting_rep,
        workspace=workspace,
        sub_indices=all_indices,
    )

    for m in meshes:
        m.identifier = f"sub representation {get_obj_identifier(energyml_object)} of {m.identifier}"

    return meshes


# MESH FILES


def _recompute_min_max(
    old_min: List,  # out parameters
    old_max: List,  # out parameters
    potential_min: List,
    potential_max: List,
) -> None:
    for i in range(len(potential_min)):
        if i >= len(old_min):
            old_min.append(potential_min[i])
        elif potential_min[i] is not None:
            old_min[i] = min(old_min[i], potential_min[i])

    for i in range(len(potential_max)):
        if i >= len(old_max):
            old_max.append(potential_max[i])
        elif potential_max[i] is not None:
            old_max[i] = max(old_max[i], potential_max[i])


def _recompute_min_max_from_points(
    old_min: List,  # out parameters
    old_max: List,  # out parameters
    points: Union[List[Point], Point],
) -> None:
    if len(points) > 0:
        if isinstance(points[0], list):
            for p in points:
                _recompute_min_max_from_points(old_min, old_max, p)
        else:
            _recompute_min_max(old_min, old_max, points, points)


def _create_shape(
    geo_type: GeoJsonGeometryType,
    point_list: List[List[float]],
    indices: Optional[Union[List[List[int]], List[int]]] = None,
    point_offset: int = 0,
    logger: Optional[Any] = None,
) -> Tuple[List, List[float], List[float]]:
    """
    Creates a shape from a point list [ [x0, y0 (, z0)? ], ..., [xn, yn (, zn)? ] ]
    using indices. If indices is a simple list, result will be a line like :  [p0, ..., pn]. With p0 and pn
    a list of coordinate from "points" parameter (like [x0, y0 (, z0)? ])
    If the indices are a list of list, result will be polygones like :
    [
        [poly0_p0, ..., poly0_pn],
        ...
        [polyn_p0, ..., polyn_pn],
    ]
    :return shape, minXYZ (as list), maxXYZ (as list)
    """
    mins = []
    maxs = []
    result = None
    try:
        if geo_type == GeoJsonGeometryType.LineString:
            result = []
            if indices is not None and len(indices) > 0:
                for idx in indices:
                    result.append(point_list[idx + point_offset])
                    _recompute_min_max_from_points(mins, maxs, point_list[idx + point_offset])
            else:
                result = point_list
                _recompute_min_max_from_points(mins, maxs, result)
        elif geo_type == GeoJsonGeometryType.MultiPoint or geo_type == GeoJsonGeometryType.Point:
            result = point_list
            _recompute_min_max_from_points(mins, maxs, result)
        elif geo_type == GeoJsonGeometryType.MultiLineString:
            if indices is not None and len(indices) > 0 and isinstance(indices[0], list):
                result = []
                for idx in indices:
                    _res, _min, _max = _create_shape(
                        geo_type=GeoJsonGeometryType.MultiLineString,
                        point_list=point_list,
                        indices=idx,
                        point_offset=point_offset,
                        logger=logger,
                    )
                    result = result + _res
                    _recompute_min_max(mins, maxs, _min, _max)
            else:
                _res, _min, _max = _create_shape(
                    geo_type=GeoJsonGeometryType.LineString,
                    point_list=point_list,
                    indices=indices,
                    point_offset=point_offset,
                    logger=logger,
                )
                result = [_res]
                _recompute_min_max(mins, maxs, _min, _max)
        elif geo_type == GeoJsonGeometryType.Polygon:
            result, mins, maxs = _create_shape(
                geo_type=GeoJsonGeometryType.MultiLineString,  # Here we only provide 1 line, the external one (outer-ring)
                point_list=point_list,
                indices=indices,
                point_offset=point_offset,
                logger=logger,
            )
            # First and last must be the same
            if len(result) > 0 and result[0] != result[-1]:
                result.append(result[0])
        elif geo_type == GeoJsonGeometryType.MultiPolygon:
            if indices is not None and len(indices) > 0 and isinstance(indices[0], list):
                result = []
                for idx in indices:
                    _res, _min, _max = _create_shape(
                        geo_type=GeoJsonGeometryType.MultiPolygon,  # Here we only provide 1 line, the external one (outer-ring)
                        point_list=point_list,
                        indices=idx,
                        point_offset=point_offset,
                        logger=logger,
                    )
                    result = result + _res
                    _recompute_min_max(mins, maxs, _min, _max)
            else:
                _res, _min, _max = _create_shape(
                    geo_type=GeoJsonGeometryType.Polygon,  # Here we only provide 1 line, the external one (outer-ring)
                    point_list=point_list,
                    indices=indices,
                    point_offset=point_offset,
                    logger=logger,
                )
                result = [_res]
                _recompute_min_max(mins, maxs, _min, _max)
    except Exception as e:
        if logger is not None:
            logger.error(e)
        # raise e
    return result, mins, maxs


def _write_geojson_shape(
    out: BytesIO,
    geo_type: GeoJsonGeometryType,
    point_list: List[List[float]],
    indices: Optional[Union[List[List[int]], List[int]]] = None,
    point_offset: int = 0,
    logger: Optional[Any] = None,
    _print_list_boundaries: Optional[bool] = True,
) -> Tuple[List[float], List[float]]:
    """
    Write a shape from a point list [ [x0, y0 (, z0)? ], ..., [xn, yn (, zn)? ] ]
    using indices. If indices is a simple list, result will be a line like :  [p0, ..., pn]. With p0 and pn
    a list of coordinate from "points" parameter (like [x0, y0 (, z0)? ])
    If the indices are a list of list, result will be polygones like :
    [
        [poly0_p0, ..., poly0_pn],
        ...
        [polyn_p0, ..., polyn_pn],
    ]
    :return shape, minXYZ (as list), maxXYZ (as list)
    """
    mins = []
    maxs = []
    try:
        if geo_type == GeoJsonGeometryType.LineString:
            if indices is not None and len(indices) > 0:
                cpt = 0
                if _print_list_boundaries:
                    out.write(b"[")
                for idx in indices:
                    out.write(json.dumps(point_list[idx + point_offset]).encode("utf-8"))
                    if cpt < len(indices) - 1:
                        out.write(b", ")
                    cpt += 1
                    _recompute_min_max_from_points(mins, maxs, point_list[idx + point_offset])
                if _print_list_boundaries:
                    out.write(b"]")
            else:
                out.write(json.dumps(point_list).encode("utf-8"))
                _recompute_min_max_from_points(mins, maxs, point_list)
        elif geo_type == GeoJsonGeometryType.MultiPoint or geo_type == GeoJsonGeometryType.Point:
            out.write(json.dumps(point_list).encode("utf-8"))
            _recompute_min_max_from_points(mins, maxs, point_list)
        elif geo_type == GeoJsonGeometryType.MultiLineString:
            if indices is not None and len(indices) > 0 and isinstance(indices[0], list):
                if _print_list_boundaries:
                    out.write(b"[")
                cpt = 0
                for idx in indices:
                    _min, _max = _write_geojson_shape(
                        out=out,
                        geo_type=GeoJsonGeometryType.MultiLineString,
                        point_list=point_list,
                        indices=idx,
                        point_offset=point_offset,
                        logger=logger,
                        _print_list_boundaries=False,
                    )
                    if cpt < len(indices) - 1:
                        out.write(b", ")
                    cpt += 1
                    _recompute_min_max(mins, maxs, _min, _max)
                if _print_list_boundaries:
                    out.write(b"]")
            else:
                if _print_list_boundaries:
                    out.write(b"[")
                _min, _max = _write_geojson_shape(
                    out=out,
                    geo_type=GeoJsonGeometryType.LineString,
                    point_list=point_list,
                    indices=indices,
                    point_offset=point_offset,
                    logger=logger,
                )
                _recompute_min_max(mins, maxs, _min, _max)
                if _print_list_boundaries:
                    out.write(b"]")
        elif geo_type == GeoJsonGeometryType.Polygon:
            # First and last must be the same
            if indices is not None and len(indices) > 0:
                if indices[0] != indices[-1]:
                    indices.append(indices[0])
            elif point_list[0] != point_list[-1]:
                point_list.append(point_list[0])

            mins, maxs = _write_geojson_shape(
                out=out,
                geo_type=GeoJsonGeometryType.MultiLineString,  # Here we only provide 1 line, the external one (outer-ring)
                point_list=point_list,
                indices=indices,
                point_offset=point_offset,
                logger=logger,
                _print_list_boundaries=_print_list_boundaries,
            )
        elif geo_type == GeoJsonGeometryType.MultiPolygon:
            if indices is not None and len(indices) > 0 and isinstance(indices[0], list):
                if _print_list_boundaries:
                    out.write(b"[")
                cpt = 0
                for idx in indices:
                    _min, _max = _write_geojson_shape(
                        out=out,
                        geo_type=GeoJsonGeometryType.MultiPolygon,  # Here we only provide 1 line, the external one (outer-ring)
                        point_list=point_list,
                        indices=idx,
                        point_offset=point_offset,
                        logger=logger,
                        _print_list_boundaries=False,
                    )
                    if cpt < len(indices) - 1:
                        out.write(b", ")
                    cpt += 1
                    _recompute_min_max(mins, maxs, _min, _max)
                if _print_list_boundaries:
                    out.write(b"]")
            else:
                if _print_list_boundaries:
                    out.write(b"[")
                _min, _max = _write_geojson_shape(
                    out=out,
                    geo_type=GeoJsonGeometryType.Polygon,  # Here we only provide 1 line, the external one (outer-ring)
                    point_list=point_list,
                    indices=indices,
                    point_offset=point_offset,
                    logger=logger,
                )
                _recompute_min_max(mins, maxs, _min, _max)
                if _print_list_boundaries:
                    out.write(b"]")
    except Exception as e:
        if logger is not None:
            logger.error(e)
        # raise e
    return mins, maxs


def to_geojson_feature(
    mesh: AbstractMesh,
    geo_type: GeoJsonGeometryType = GeoJsonGeometryType.Point,
    geo_type_prefix: Optional[str] = "AnyCrs",
    properties: Optional[dict] = None,
    point_offset: int = 0,
    logger=None,
) -> Dict:
    feature = {}

    if mesh.point_list is not None and len(mesh.point_list) > 0:
        points = mesh.point_list

        #  TODO: remove :
        # points = list(map(
        #     lambda p: list(map(lambda x: round(x/10000., 4), p)),
        #     mesh.point_list
        # ))

        indices = mesh.get_indices()
        # polygon must have the first and last point as the same
        if geo_type == GeoJsonGeometryType.Polygon or geo_type == GeoJsonGeometryType.MultiPolygon:
            if logger is not None:
                logger.debug("# to_geojson_feature > Reshaping indices for polygons")
            if indices is not None:
                for indices_i in indices:
                    indices_i.append(indices_i[0])
            if logger is not None:
                logger.debug("\t# to_geojson_feature > Indices reshaped")

        if logger is not None:
            logger.debug("# to_geojson_feature > Computing shape")

        coordinates, mins, maxs = _create_shape(
            geo_type=geo_type,
            point_list=points,
            indices=indices,
            point_offset=point_offset,
            logger=logger,
        )

        # Pop previously added last :
        if geo_type == GeoJsonGeometryType.Polygon or geo_type == GeoJsonGeometryType.MultiPolygon:
            if indices is not None:
                for indices_i in indices:
                    indices_i.pop()

        if logger is not None:
            logger.debug("\t# to_geojson_feature > shaped")

        bbox_geometry = []  # TODO : see : https://www.rfc-editor.org/rfc/rfc7946#section-5

        bbox_geometry = mins + maxs

        geometry = {
            # "type": f"{geo_type_prefix}{geo_type.name}",
            "type": f"{geo_type.name}",
            "coordinates": coordinates,
            "bbox": bbox_geometry,
        }

        feature = {
            "type": f"{geo_type_prefix}Feature",
            "properties": properties or {},
            "geometry": geometry,
        }

    return feature


def write_geojson_feature(
    out: BytesIO,
    mesh: AbstractMesh,
    geo_type: GeoJsonGeometryType = GeoJsonGeometryType.Point,
    geo_type_prefix: Optional[str] = "AnyCrs",
    properties: Optional[dict] = None,
    point_offset: int = 0,
    logger=None,
) -> None:
    if mesh.point_list is not None and len(mesh.point_list) > 0:
        points = mesh.point_list

        indices = mesh.get_indices()
        # polygon must have the first and last point as the same
        if geo_type == GeoJsonGeometryType.Polygon or geo_type == GeoJsonGeometryType.MultiPolygon:
            if logger is not None:
                logger.debug("# to_geojson_feature > Reshaping indices for polygons")
            if indices is not None:
                for indices_i in indices:
                    indices_i.append(indices_i[0])
            if logger is not None:
                logger.debug("\t# to_geojson_feature > Indices reshaped")

        if logger is not None:
            logger.debug("# to_geojson_feature > Computing shape")

        out.write(b"{")  # start feature
        out.write(f'"type": "{geo_type_prefix}Feature", '.encode())
        out.write(f'"properties": {json.dumps(properties or {}) }, '.encode())
        out.write(b'"geometry": ')

        out.write(b"{")  # start geometry
        # "type": f"{geo_type_prefix}{geo_type.name}",
        out.write(f'"type": "{geo_type.name}", '.encode())
        out.write(f'"coordinates": '.encode())
        mins, maxs = _write_geojson_shape(
            out=out,
            geo_type=geo_type,
            point_list=points,
            indices=indices,
            point_offset=point_offset,
            logger=logger,
        )
        bbox_geometry = mins + maxs  # TODO : see : https://www.rfc-editor.org/rfc/rfc7946#section-5

        out.write(f', "bbox": {json.dumps(bbox_geometry)}'.encode())
        out.write(b"}")  # end geometry

        # Pop previously added last :
        if geo_type == GeoJsonGeometryType.Polygon or geo_type == GeoJsonGeometryType.MultiPolygon:
            if indices is not None:
                for indices_i in indices:
                    indices_i.pop()

        if logger is not None:
            logger.debug("\t# to_geojson_feature > shaped")

        out.write(b"}")  # End feature


def mesh_to_geojson_type(obj: AbstractMesh) -> GeoJsonGeometryType:
    if isinstance(obj, SurfaceMesh):
        return GeoJsonGeometryType.MultiPolygon
    elif isinstance(obj, PolylineSetMesh):
        return GeoJsonGeometryType.MultiLineString
    else:
        return GeoJsonGeometryType.MultiPoint


def export_geojson_io(
    out: BytesIO,
    mesh_list: List[AbstractMesh],
    obj_name: Optional[str] = None,
    properties: Optional[List[Optional[Dict]]] = None,
    global_properties: Optional[Dict] = None,
    logger: Optional[Any] = None,
):
    out.write(b"{")
    out.write(b'"type": "FeatureCollection",')
    if obj_name is not None:
        out.write(b'"name": "')
        out.write(obj_name.encode())
        out.write(b'",')

    if global_properties is not None and len(global_properties) > 0:
        for k, v in global_properties.items():
            out.write(b'"')
            out.write(k.encode())
            out.write(b'": ')
            out.write(json.dumps(v).encode())
            out.write(b",")

    out.write(b'"features": [')

    cpt = 0
    point_offset = 0

    for mesh in mesh_list:
        pos = out.tell()
        write_geojson_feature(
            out=out,
            mesh=mesh,
            geo_type=mesh_to_geojson_type(mesh),
            properties=properties[cpt] if properties is not None and len(properties) > cpt else None,
            point_offset=0,  # point_offset,
            logger=logger,
        )
        if out.tell() != pos and cpt < len(mesh_list) - 1:
            out.write(b",")
        cpt += 1
        point_offset = point_offset + len(mesh.point_list)
    out.write(b"]")  # end features
    out.write(b"}")  # end geojson


def export_geojson_dict(
    mesh_list: List[AbstractMesh],
    obj_name: Optional[str] = None,
    properties: Optional[List[Optional[Dict]]] = None,
    logger: Optional[Any] = None,
):
    res = {"type": "FeatureCollection", "features": []}
    cpt = 0
    point_offset = 0
    for mesh in mesh_list:
        feature = to_geojson_feature(
            mesh=mesh,
            geo_type=mesh_to_geojson_type(mesh),
            properties=properties[cpt] if properties is not None and len(properties) > cpt else None,
            point_offset=0,  # point_offset,
            logger=logger,
        )
        if feature is not None:
            res["features"].append(feature)
        cpt += 1
        point_offset = point_offset + len(mesh.point_list)

    return res


def export_off(mesh_list: List[AbstractMesh], out: BytesIO):
    """
    Export an :class:`AbstractMesh` into off format.
    :param mesh_list:
    :param out:
    :return:
    """
    nb_points = sum(list(map(lambda m: len(m.point_list), mesh_list)))
    nb_edges = sum(list(map(lambda m: m.get_nb_edges(), mesh_list)))
    nb_faces = sum(list(map(lambda m: m.get_nb_faces(), mesh_list)))

    out.write(b"OFF\n")
    out.write(_FILE_HEADER)
    out.write(f"{nb_points} {nb_faces} {nb_edges}\n".encode("utf-8"))

    points_io = BytesIO()
    faces_io = BytesIO()

    point_offset = 0
    for m in mesh_list:
        export_off_part(
            off_point_part=points_io,
            off_face_part=faces_io,
            points=m.point_list,
            indices=m.get_indices(),
            point_offset=point_offset,
            colors=[],
        )
        point_offset = point_offset + len(m.point_list)

    out.write(points_io.getbuffer())
    out.write(faces_io.getbuffer())


def export_off_part(
    off_point_part: BytesIO,
    off_face_part: BytesIO,
    points: List[List[float]],
    indices: List[List[int]],
    point_offset: Optional[int] = 0,
    colors: Optional[List[List[int]]] = None,
) -> None:
    for p in points:
        for pi in p:
            off_point_part.write(f"{pi} ".encode("utf-8"))
        off_point_part.write(b"\n")

    cpt = 0
    for face in indices:
        if len(face) > 1:
            off_face_part.write(f"{len(face)} ".encode("utf-8"))
            for pi in face:
                off_face_part.write(f"{pi + point_offset} ".encode("utf-8"))

            if colors is not None and len(colors) > cpt and colors[cpt] is not None and len(colors[cpt]) > 0:
                for col in colors[cpt]:
                    off_face_part.write(f"{col} ".encode("utf-8"))

            off_face_part.write(b"\n")
        cpt += 1


def export_obj(mesh_list: List[AbstractMesh], out: BytesIO, obj_name: Optional[str] = None):
    """
    Export an :class:`AbstractMesh` into obj format.

    Each AbstractMesh from the list :param:`mesh_list` will be placed into its own group.
    :param mesh_list:
    :param out:
    :param obj_name:
    :return:
    """
    out.write("# Generated by energyml-utils a Geosiris python module\n\n".encode("utf-8"))

    if obj_name is not None:
        out.write(f"o {obj_name}\n\n".encode("utf-8"))

    point_offset = 0
    for m in mesh_list:
        out.write(f"g {m.identifier}\n\n".encode("utf-8"))
        _export_obj_elt(
            off_point_part=out,
            off_face_part=out,
            points=m.point_list,
            indices=m.get_indices(),
            point_offset=point_offset,
            colors=[],
            elt_letter="l" if isinstance(m, PolylineSetMesh) else "f",
        )
        point_offset = point_offset + len(m.point_list)
        out.write("\n".encode("utf-8"))


def _export_obj_elt(
    off_point_part: BytesIO,
    off_face_part: BytesIO,
    points: List[List[float]],
    indices: List[List[int]],
    point_offset: Optional[int] = 0,
    colors: Optional[List[List[int]]] = None,
    elt_letter: str = "f",
) -> None:
    """

    :param off_point_part:
    :param off_face_part:
    :param points:
    :param indices:
    :param point_offset:
    :param colors: currently not supported
    :param elt_letter: "l" for line and "f" for faces
    :return:
    """
    offset_obj = 1  # OBJ point indices starts at 1 not 0
    for p in points:
        if len(p) > 0:
            off_point_part.write(f"v {' '.join(list(map(lambda xyz: str(xyz), p)))}\n".encode("utf-8"))

    # cpt = 0
    for face in indices:
        if len(face) > 1:
            off_face_part.write(
                f"{elt_letter} {' '.join(list(map(lambda x: str(x + point_offset + offset_obj), face)))}\n".encode(
                    "utf-8"
                )
            )

            # if colors is not None and len(colors) > cpt and colors[cpt] is not None and len(colors[cpt]) > 0:
            #     for col in colors[cpt]:
            #         off_face_part.write(f"{col} ".encode('utf-8'))

            # off_face_part.write(b"\n")


def export_multiple_data(
    epc_path: str,
    uuid_list: List[str],
    output_folder_path: str,
    output_file_path_suffix: str = "",
    file_format: MeshFileFormat = MeshFileFormat.OBJ,
    use_crs_displacement: bool = True,
    logger: Optional[Any] = None,
):
    epc = Epc.read_file(epc_path)

    # with open(epc_path.replace(".epc", ".h5"), "rb") as fh:
    #     buf = BytesIO(fh.read())
    #     epc.h5_io_files.append(buf)

    try:
        os.makedirs(output_folder_path, exist_ok=True)
    except OSError:
        pass

    for uuid in uuid_list:
        energyml_obj = None
        try:
            energyml_obj = epc.get_object_by_uuid(uuid)[0]
        except:
            if logger is not None:
                logger.error(f"Object with uuid {uuid} not found")
            else:
                logging.error(f"Object with uuid {uuid} not found")
            continue
        file_name = (
            f"{gen_energyml_object_path(energyml_obj)}_"
            f"[{get_object_attribute(energyml_obj, 'citation.title')}]"
            f"{output_file_path_suffix}"
            f".{file_format.value}"
        )
        file_path = f"{output_folder_path}/{file_name}"
        logging.debug(f"Exporting : {file_path}")
        mesh_list = read_mesh_object(
            energyml_object=energyml_obj,
            workspace=epc,
            use_crs_displacement=use_crs_displacement,
        )
        if file_format == MeshFileFormat.OBJ:
            with open(file_path, "wb") as f:
                export_obj(
                    mesh_list=mesh_list,
                    out=f,
                )
        elif file_format == MeshFileFormat.OFF:
            with open(file_path, "wb") as f:
                export_off(
                    mesh_list=mesh_list,
                    out=f,
                )
        elif file_format == MeshFileFormat.GEOJSON:
            with open(file_path, "wb") as f:
                export_geojson_io(
                    out=f,
                    mesh_list=mesh_list,
                    logger=logger,
                    global_properties={"epc_path": epc_path},
                )
        else:
            logging.error(f"Code is not written for format {file_format}")
